# -*- coding: utf-8 -*-
"""
Created on Mon Jan 27 17:21:13 2020

@author: veenstra

"""

import pytest
import os
import numpy as np
import pandas as pd
import datetime as dt
import hatyan
import pytz

dir_tests = os.path.dirname(__file__) #F9 doesnt work, only F5 (F5 also only method to reload external definition scripts)
dir_testdata = os.path.join(dir_tests,'data_unitsystemtests')


@pytest.mark.systemtest
def test_frommergedcomp():
    # 1. define test data
    current_station = 'VLISSGN'
    
    #comp
    file_data_comp = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
    
    #pred
    times_pred = slice(dt.datetime(2019,1,1),dt.datetime(2019,1,1,12), "10min")
    
    # 2. define initial expectations
    expected_ts_prediction_data_pd_values = np.array([1.00809295,  0.89565827,  0.77557688,  0.64880508,  0.51669927,
                                                      0.38083644,  0.24282757,  0.1041679 , -0.03384444, -0.17011025,
                                                     -0.3036618 , -0.43355561, -0.55876447, -0.67810779, -0.79024821,
                                                     -0.89376351, -0.98728197, -1.06965037, -1.14009357, -1.19832413,
                                                     -1.2445716 , -1.27952004, -1.30416467, -1.31961912, -1.32691704,
                                                     -1.32685401, -1.31990568, -1.30623924, -1.28581172, -1.25852648,
                                                     -1.22440367, -1.18371671, -1.1370541 , -1.08528461, -1.02942801,
                                                     -0.97045777, -0.90908055, -0.84554483, -0.77952619, -0.71012042,
                                                     -0.63595176, -0.55537797, -0.46675305, -0.36869694, -0.26032191,
                                                     -0.14137807, -0.01230058,  0.12583519,  0.27142446,  0.4224715 ,
                                                      0.57671752,  0.73172934,  0.88493925,  1.03364646,  1.17500835,
                                                      1.30605751,  1.42377775,  1.52525728,  1.60791471,  1.66976848,
                                                      1.70970034,  1.72765351,  1.72471102,  1.70301805,  1.66554205,
                                                      1.61569821,  1.55689834,  1.49210117,  1.42344542,  1.35203362,
                                                      1.27790547,  1.20020227,  1.11748521])
    
    # 3. run test
    COMP_mergedfromfile = hatyan.read_components(filename=file_data_comp)
    ts_prediction_direct = hatyan.prediction(COMP_mergedfromfile, times=times_pred)
    ts_prediction_direct_values = ts_prediction_direct['values'].values
    
    # 4. Vefiry final expectations
    assert type(ts_prediction_direct) == pd.core.frame.DataFrame
    assert ts_prediction_direct.index.tz==pytz.FixedOffset(60)
    assert ts_prediction_direct.index[0].tz_localize(None) == times_pred.start
    assert ts_prediction_direct.index[-1].tz_localize(None) == times_pred.stop
    assert len(ts_prediction_direct_values) == len(expected_ts_prediction_data_pd_values)
    assert type(ts_prediction_direct_values) == type(expected_ts_prediction_data_pd_values)
    assert (np.abs(ts_prediction_direct_values - expected_ts_prediction_data_pd_values) < 10E-9).all()


@pytest.mark.systemtest
def test_meas_HWLW_toomuch():
    """
    this test will fail if the minimal prominence is set to 0.01 or lower, or of the minimal width=None
    Then there will be an additional LW found right after the HW. With a slightly higher minimal prominence (>0.02) this is avoided.
    This testcase originates from an issue with a FEWS timeseries: https://github.com/Deltares/hatyan/issues/85
    """
    wl_vals = np.array([ 0.4437,  0.563 ,  0.6677,  0.7498,  0.8125,  0.8732,  0.9181,
            0.9384,  0.944 ,  0.94  ,  0.9254,  0.9033,  0.8602,  0.8258,
            0.8177,  0.7867,  0.7184,  0.6672,  0.6314,  0.5722,  0.5028,
            0.428 ,  0.3539,  0.2822,  0.2003,  0.1141,  0.0317, -0.0457,
           -0.1233, -0.2187, -0.3101, -0.3945, -0.476 , -0.5503, -0.6137,
           -0.671 , -0.7262, -0.7742, -0.8025, -0.8145, -0.8271, -0.829 ,
           -0.8131, -0.7856, -0.7616, -0.7289, -0.7176, -0.7114, -0.7088,
           -0.7111, -0.7113, -0.7106, -0.715 , -0.7282, -0.7392, -0.7434,
           -0.7384, -0.7405, -0.7321, -0.7202, -0.7054, -0.6847, -0.6514,
           -0.6082, -0.5656, -0.5155, -0.4549, -0.3814, -0.2934, -0.1994,
           -0.0949,  0.0291,  0.1794,  0.3537,  0.5279,  0.6908,  0.8318,
            0.9402,  1.024 ,  1.0951,  1.1474,  1.1639,  1.1582,  1.147 ,
            1.1298,  1.0997,  1.0717,  1.0828,  1.078 ,  1.0232,  0.9849,
            0.969 ,  0.9424,  0.9058,  0.8601,  0.8045,  0.7569,  0.6917,
            0.6198,  0.5531,  0.4802,  0.4109,  0.3291,  0.2442,  0.1692,
            0.0838, -0.0073, -0.088 , -0.1659, -0.2412, -0.3055, -0.3647,
           -0.405 , -0.4294, -0.4459, -0.4414, -0.4359, -0.4387, -0.4455,
           -0.4544, -0.4641, -0.4793, -0.4984, -0.5001, -0.4996, -0.5065,
           -0.5022, -0.4981, -0.5046, -0.5136, -0.5243, -0.526 , -0.5254,
           -0.5239, -0.5131, -0.4995, -0.4831, -0.4595, -0.4368, -0.4089,
           -0.3758, -0.3415, -0.299 , -0.2475, -0.1889, -0.11  , -0.0254,
            0.0702,  0.1958,  0.3341,  0.4824,  0.6129,  0.709 ,  0.7898,
            0.8648,  0.9185,  0.955 ,  0.9721,  0.971 ,  0.9652,  0.9486,
            0.9194,  0.8885,  0.8618,  0.8491,  0.8154,  0.7478,  0.7017,
            0.6724,  0.6105,  0.5356,  0.4669,  0.3918,  0.323 ,  0.2405])
    wl_times = pd.date_range('2023-06-13 21:00','2023-06-15 02:00',freq='10min')

    # Then store it as a pandas dataframe and compute extremes
    wl_pd = pd.DataFrame({'values':wl_vals},index=wl_times)
    wl_pd_ext = hatyan.calc_HWLW(wl_pd)

    assert len(wl_pd_ext) == 3
    assert (wl_pd_ext['HWLWcode'] == [2,1,2]).all()


@pytest.mark.systemtest
def test_meas_HWLW_toomuch_19y():
    """
    Additional testcase for https://github.com/Deltares/hatyan/issues/85
    The min_width parameter for calc_HWLW() has to be approx 2hrs to properly compute and number almost all of the 19y timeseries
    Stations HELLVSS/KRIMPADLK/RAKND still fail, but it seems arbitrary and sometimes computing+numbering extremes per year does work
    """
    current_station = 'HOEKVHLD'
    file_dia = os.path.join(dir_testdata,f'{current_station}_obs19.txt')
    
    wl_pd = hatyan.read_dia(file_dia)
    
    wl_pd_ext = hatyan.calc_HWLW(wl_pd)
    
    # For testing: plot water level timeseries with peaks identified and labeled. 
    #hatyan.plot_timeseries(ts=wl_pd,ts_ext=wl_pd_ext)
    
    # RUN HATYAN: Assign numbers to the extremes
    wl_pd_ext = hatyan.calc_HWLWnumbering(wl_pd_ext,station=current_station)

    
@pytest.mark.systemtest
def test_calc_HWLW_ts_with_gap():
    """
    testcase that checks if no extreme is computed at the start/end of a gap
    like was documented in https://github.com/Deltares/hatyan/issues/97
    """
    
    current_station = 'VLISSGN'
    
    file_pred = os.path.join(dir_testdata,f'{current_station}_pre.txt')
    ts_prediction = hatyan.read_dia(filename=file_pred, station=current_station)
    ts_prediction = ts_prediction.loc[slice("2019-01","2019-01")]
    ts_prediction.loc[slice("2019-01-12 04:00", "2019-01-19 04:00")] = np.nan
    
    ts_ext_prediction = hatyan.calc_HWLW(ts=ts_prediction)
    
    # assert the number of extremes, was 91 before the fix of the issue
    assert len(ts_ext_prediction) == 90


@pytest.mark.systemtest
def test_frommergedcomp_HWLW_toomuch():
    """
    This test produces a very short prediction for DENHDR, based on an imported component list. It then calculates extremes (HW/LW) and numbers them both.
    This test is meant to tweak the prominence parameter of scipy.signal.find_peaks() in hatyan.calc_HWLW() and it fails when the prominence is not set or is too small.
    A prominence value of None works for most stations, but fails for DENHDR in this period since a local dip is interpreted as a low water.
    This incorrect LW has a prominence of ~0.03, other peaks in this timeseries have a prominence of >1 and even when lowering M2 to an unrealistic value (0.5*M2) the prominence of peaks is >0.2
    A prominence value of 0.1 makes sure this test succeeds.
    
    """
    
    #for testing occurance of invalid low water at start of denhelder timeseries
    times_pred = slice(dt.datetime(2009,12,31,14),dt.datetime(2010,1,2,12), "1min")
    current_station = 'DENHDR'

    file_data_comp0 = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
    comp_merged = hatyan.read_components(filename=file_data_comp0)
    
    ts_prediction_HWLWno = hatyan.prediction(comp=comp_merged, times=times_pred)
    ts_ext_prediction_HWLWno_pre = hatyan.calc_HWLW(ts=ts_prediction_HWLWno)
    
    ts_ext_prediction_HWLWno = hatyan.calc_HWLWnumbering(ts_ext=ts_ext_prediction_HWLWno_pre, station=current_station)
    
    values_actual = ts_ext_prediction_HWLWno_pre['values'].values
    values_expected = np.array([-0.80339617,  0.61842454, -0.76465349,  0.79758517, -0.87872312])
    hwlwcodes_actual = ts_ext_prediction_HWLWno_pre['HWLWcode'].values
    hwlwcodes_expected = np.array([2, 1, 2, 1, 2])
    hwlwnos_actual = ts_ext_prediction_HWLWno['HWLWno'].values
    hwlwnos_expected = np.array([7057, 7058, 7058, 7059, 7059])
    
    assert len(ts_ext_prediction_HWLWno_pre) == 5
    assert np.allclose(values_actual, values_expected)
    assert np.allclose(hwlwcodes_actual, hwlwcodes_expected)
    assert np.allclose(hwlwnos_actual, hwlwnos_expected)


@pytest.mark.systemtest
#@pytest.mark.parametrize("current_station, yr", [pytest.param(x, y, id='%s %d'%(x,y))  for y in range(1999,2022) for x in ['WICK','ABDN','LEITH','WHITBY','IMMHM','CROMR','FELSWE','CADZD','VLISSGN','TERNZN','ROOMPBTN','HARVT10','HOEKVHLD','ROTTDM','DORDT','SCHEVNGN','IJMDBTHVN','PETTZD','DENHDR','DENOVBTN','HARLGN','HOLWD','SCHIERMNOG','LAUWOG','EEMSHVN','DELFZL','CUXHVN']])
@pytest.mark.parametrize("current_station, yr", [pytest.param(x, y, id='%s_%d'%(x,y)) for y in [2018,2022] for x in ['HOEKVHLD','ROTTDM','DENHDR','LITHDP']])
def test_frommergedcomp_HWLW_toolittle(current_station, yr):
    """
    This test produces a prediction for an entire year for a specific station, based on an imported component list. It then calculates extremes (HW/LW) and numbers them both. 
    The test fails if the numbers for HW and/or LW do not always increase with 1 and prints parts of the HW/LW-arrays with these gaps.
    This test is meant to tweak the distance parameter of scipy.signal.find_peaks() in hatyan.calc_HWLW() and it fails when the distance is to small.
    A distance value of M2period/1.4 works for most stations, but fails for three stations of the selection in several years (of which the year 2000)
    A distance value of M2period/1.5 for HW and M2period/1.7 for LW works for all tested stations and years.
    
    #selected stations which previously resulted in missing HWLW values in year 2000 and 2018 (and more)
    current_station = 'HOEKVHLD'
    current_station = 'ROTTDM'
    current_station = 'DENHDR'
    current_station = 'LITHDP' #still fails, but is accepted
    yr=2000
    yr=2018
    """
    
    #stats_all = ['ABDN','AMLAHVN','BAALHK','BATH','BERGSDSWT','BORSSLE','BOURNMH','BRESKS','BROUWHVSGT02','BROUWHVSGT08','CADZD','CROMR','CUXHVN','DELFZL','DENHDR','DENOVBTN','DEVPT','DORDT','DOVR','EEMHVN','EEMSHVN','EURPFM','EURPHVN','FELSWE','FISHGD','GEULHVN','GOIDSOD','GOUDBG','HAGSBNDN','HANSWT','HARLGN','HARMSBG','HARTBG','HARTHVN','HARVT10','HEESBN','HELLVSS','HOEKVHLD','HOLWD','HUIBGT','IJMDBTHVN','IJMDSMPL','IMMHM','KATSBTN','KEIZVR','KINLBVE','KORNWDZBTN','KRAMMSZWT','KRIMPADIJSL','KRIMPADLK','K13APFM','LAUWOG','LEITH','LICHTELGRE','LITHDP','LLANDNO','LOWST','MAASMSMPL','MAASSS','MAESLKRZZDE','MARLGT','MOERDK','NES','NEWHVN','NEWLN','NIEUWSTZL','NORTHSS','OOSTSDE04','OOSTSDE11','OOSTSDE14','OUDSD','OVLVHWT','PARKSS','PETTZD','PORTSMH','RAKND','ROOMPBNN','ROOMPBTN','ROTTDM','ROZBSSNZDE','ROZBSSZZDE','SCHAARVDND','SCHEVNGN','SCHIERMNOG','SCHOONHVN','SHEERNS','SINTANLHVSGR','SPIJKNSE','STAVNSE','STELLDBTN','STORNWY','SUURHBNZDE','TENNSHVN','TERNZN','TERSLNZE','TEXNZE','VLAARDGN','VLAKTVDRN','VLIELHVN','VLISSGN','VURN','WALSODN','WERKDBTN','WESTKPLE','WESTTSLG','WEYMH','WHITBY','WICK','WIERMGDN','YERSKE','ZALTBML','A12','AUKFPFM','AWGPFM','D15','F16','F3PFM','J6','K14PFM','L9PFM','NORTHCMRT','Q1']
    stats_xfac0 = ['A12','ABDN','AUKFPFM','BOURNMH','CROMR','CUXHVN','D15','DEVPT','DOVR','F16','F3PFM','FELSWE','FISHGD','IMMHM','J6','K13APFM','K14PFM','KINLBVE','LEITH','LLANDNO','LOWST','NEWHVN','NEWLN','NORTHCMRT','NORTHSS','PORTSMH','SHEERNS','STORNWY','WEYMH','WHITBY','WICK']
    #stats_anaperyear0 = ['A12','ABDN','AUKFPFM','BOURNMH','CROMR','D15','DEVPT','DOVR','F16','F3PFM','FELSWE','FISHGD','IMMHM','J6','K14PFM','KINLBVE','LEITH','LLANDNO','LOWST','NEWHVN','NEWLN','NORTHCMRT','NORTHSS','PORTSMH','SHEERNS','STORNWY','WEYMH','WHITBY','WICK']
    #stats_MSL = ['EURPFM','K13APFM','LICHTELGRE','A12','AUKFPFM','AWGPFM','D15','F16','F3PFM','J6','K14PFM','L9PFM','NORTHCMRT','Q1']
    
    #selected_stations = stats_all
    #selected_stations = ['WICK','ABDN','LEITH','WHITBY','IMMHM','CROMR','FELSWE','CADZD','VLISSGN','TERNZN','ROOMPBTN','HARVT10','HOEKVHLD','ROTTDM','DORDT','SCHEVNGN','IJMDBTHVN','PETTZD','DENHDR','DENOVBTN','HARLGN','HOLWD','SCHIERMNOG','LAUWOG','EEMSHVN','DELFZL','CUXHVN']
    #selected_stations = ['CROMR','CADZD','HOEKVHLD','DENHDR','CUXHVN']
    #selected_stations = ['HOEKVHLD','ROTTDM','DENHDR'] #selected stations which resulted in missing HWLW values 
    #for yr in [2000]: #range(1999,2022):
    #fig, (ax1) = plt.subplots(1,1,figsize=(15,6))
    #n_colors = len(selected_stations)
    #colors = plt.cm.jet(np.linspace(0,1,n_colors))
    #for i_stat, current_station in enumerate(selected_stations):
    print('-'*100)
    print('%-45s = %s'%('station_name',current_station))
    print('-'*5)
    
    #START OF STATION SETTINGS
    #xfactor
    if current_station in stats_xfac0:
        xfac=False
    else:
        xfac=True
    #analysis_peryear
    #if current_station in stats_anaperyear0:
    #    analysis_perperiod=False
    #else:
    #    analysis_perperiod='Y'
    #constituent list
    #const_list = hatyan.get_const_list_hatyan('year') #94 const
    #vertical reference
    #vertref='NAP'
    #END OF STATION SETTINGS
    
    file_data_comp0 = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
    times_pred = slice(dt.datetime(yr,1,1),dt.datetime(yr+1,1,1), "1min")
    
    #component groups
    COMP_merged = hatyan.read_components(filename=file_data_comp0)
    assert COMP_merged.attrs["xfac"] == xfac
    
    #prediction and validation
    ts_prediction = hatyan.prediction(comp=COMP_merged, times=times_pred)
    #ts_validation = hatyan.read_dia(filename=file_data_predvali, station=current_station)
    #ts_ext_validation = hatyan.read_dia(filename=file_data_predvaliHWLW, station=current_station)
    ts_ext_prediction = hatyan.calc_HWLW(ts=ts_prediction)
    
    #calculate tidal wave number
    ts_ext_prediction_HWLWno = hatyan.calc_HWLWnumbering(ts_ext=ts_ext_prediction, station=current_station)
    #print(ts_ext_prediction_HWLWno)
    #for irow, pdrow in ts_ext_prediction_HWLWno.iterrows():
    #    ax1.text(pdrow.index,pdrow['values'],pdrow['HWLWno'], color=colors[i_stat])
    
    #ax1.plot(ts_prediction.index, ts_prediction['values'], label=current_station, color=colors[i_stat])
    #figa, (ax1a) = hatyan.plot_timeseries(ts=ts_prediction, ts_ext=ts_ext_prediction_HWLWno)
    
    #str_combination = '%i_%s'%(yr, current_station)
    #print(str_combination)
    print(current_station, yr)
    print('all HWLW values:')
    print(ts_ext_prediction_HWLWno)
    
    HW_data = ts_ext_prediction_HWLWno[ts_ext_prediction_HWLWno['HWLWcode']==1]
    HW_data_diff1bool = (HW_data['HWLWno'].diff().iloc[1:].values==1)
    print('%i parts of HW array containing gaps:'%((~HW_data_diff1bool).sum()))
    if not HW_data_diff1bool.all():
        ids_false = np.nonzero(~HW_data_diff1bool)[0]
        for id_false in ids_false: 
            print(HW_data.iloc[id_false-1:id_false+3])
    
    LW_data = ts_ext_prediction_HWLWno[ts_ext_prediction_HWLWno['HWLWcode']==2]
    LW_data_diff1bool = (LW_data['HWLWno'].diff().iloc[1:].values==1)
    print('%i parts of LW array containing gaps:'%((~LW_data_diff1bool).sum()))
    if not LW_data_diff1bool.all():
        ids_false = np.nonzero(~LW_data_diff1bool)[0]
        for id_false in ids_false: 
            print(LW_data.iloc[id_false-1:id_false+3])
    
    assert HW_data_diff1bool.all()
    assert LW_data_diff1bool.all()
    
    #ax1.set_xlim(times_ext_pred)
    #fig.tight_layout()
    #ax1.legend(loc=2, fontsize=7)#bbox_to_anchor=(1,1))
    #import matplotlib.dates as mdates
    #ax1.xaxis.set_major_formatter(mdates.DateFormatter("%Y-%m-%d %H:%M"))
    #fig.savefig('tide_numbering_%i.png'%(yr), dpi=250)


@pytest.mark.parametrize("current_station", [pytest.param(x, id=x) for x in ['HOEKVHLD','DENHDR']])
@pytest.mark.systemtest
def test_frommergedcomp_HWLW_345(current_station):
    """
    This test produces a prediction for a period for HOEKVHLD, based on an imported component list. It then calculates extremes (HW/LW) with several settings and numbers them both (including 3/4/5 HWLWcodes around aggers, excluding 11/22 HWLWcodes). 
    #HOEKVHLD has alternating aggers, DENHDR has double HW's (which should not be numbered as aggers)
    
    current_station = 'HOEKVHLD'
    current_station = 'DENHDR'
    """
    
    file_data_comp0 = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
    comp_merged = hatyan.read_components(filename=file_data_comp0)
    
    times_pred = slice(dt.datetime(2010,1,31,3),dt.datetime(2010,2,17,12), "1min") #longer period with alternating aggers and no aggers, also eerste HW wordt als lokaal ipv primair HW gezien, also extra agger outside of 1stLW/agger/2ndLW sequence
    # times_pred = slice(dt.datetime(2010,1,31),dt.datetime(2010,2,3), "1min") #extra agger outside of 1stLW/agger/2ndLW sequence
    # times_pred = slice(dt.datetime(2010,6,26),dt.datetime(2010,6,28), "1min") #lokale extremen tussen twee laagste LWs
    # times_pred = slice(dt.datetime(2019,2,1),dt.datetime(2019,2,2), "1min") #eerste HW wordt als lokaal ipv primair HW gezien (lage prominence door dicht op begin tijdserie) >> warning
    
    ts_prediction_HWLWno = hatyan.prediction(comp=comp_merged, times=times_pred)
    ts_ext_prediction_main = hatyan.calc_HWLW(ts=ts_prediction_HWLWno)
    #ts_ext_prediction_all = hatyan.calc_HWLW(ts=ts_prediction_HWLWno, calc_HWLW345=True, calc_HWLW345_cleanup1122=False)
    ts_ext_prediction_clean = hatyan.calc_HWLW(ts=ts_prediction_HWLWno, calc_HWLW345=True)#, calc_HWLW345_cleanup1122=True) #for numbering, cannot cope with 11/22 HWLWcodes
    ts_ext_prediction_clean_tomain = hatyan.calc_HWLW12345to12(ts_ext_prediction_clean)
    
    ts_ext_prediction_main_HWLWno = hatyan.calc_HWLWnumbering(ts_ext=ts_ext_prediction_main, station=current_station)
    ts_ext_prediction_clean_HWLWno = hatyan.calc_HWLWnumbering(ts_ext=ts_ext_prediction_clean, station=current_station)
    
    #fig, (ax1,ax2) = hatyan.plot_timeseries(ts=ts_prediction_HWLWno, ts_ext=ts_ext_prediction_all)#, ts_ext_validation=ts_ext_validation)
    #for irow, pdrow in ts_ext_prediction_clean_HWLWno.iterrows():
    #    ax1.text(pdrow.index,pdrow['values'],pdrow['HWLWno'])
    
    #expected values
    if current_station == 'HOEKVHLD':
        expected_ts_ext_prediction_main_HWLWno_HWLWno = np.array([7117, 7117, 7118, 7118, 7119, 7119, 7120, 7120, 7121,
                                                                  7121, 7122, 7122, 7123, 7123, 7124, 7124, 7125, 7125, 7126, 7126,
                                                                  7127, 7127, 7128, 7128, 7129, 7129, 7130, 7130, 7131, 7131, 7132,
                                                                  7132, 7133, 7133, 7134, 7134, 7135, 7135, 7136, 7136, 7137, 7137,
                                                                  7138, 7138, 7139, 7139, 7140, 7140, 7141, 7141, 7142, 7142, 7143,
                                                                  7143, 7144, 7144, 7145, 7145, 7146, 7146, 7147, 7147, 7148, 7148,
                                                                  7149])
        expected_ts_ext_prediction_main_HWLWcode = np.array([1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2,
                                                             1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2,
                                                             1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2,
                                                             1])
        """
        expected_ts_ext_prediction_all_HWLWcode = np.array([11, 22, 11, 22,  1,  3,  4,  5,  1,  3,  4,  5,  1,  3,  4,  5,  1,
                                                             3,  4,  5,  1,  3,  4,  5,  1,  3,  4,  5,  1,  3,  4,  5,  1,  3,
                                                             4,  5,  1,  3,  4,  5,  1,  3,  4,  5,  1,  2,  1,  2,  1,  2,  1,
                                                             2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  3,  4,  5,
                                                             1,  2,  1,  3,  4,  5,  1,  2,  1,  3,  4,  5,  1,  2,  1,  3,  4,
                                                             5,  1,  2,  1,  2,  1,  3,  4,  5,  1,  3,  4,  5,  1,  3,  4,  5,
                                                             1, 22, 11])
        """
        expected_ts_ext_prediction_clean_HWLWno_HWLWcode = np.array([1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3,
                                                                     4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 2, 1, 2,
                                                                     1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 3, 4, 5, 1, 2,
                                                                     1, 3, 4, 5, 1, 2, 1, 3, 4, 5, 1, 2, 1, 3, 4, 5, 1, 2, 1, 2, 1, 3,
                                                                     4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1])
        expected_ts_ext_prediction_clean_HWLWno_HWLWno = np.array([7117, 7117, 7117, 7117, 7118, 7118, 7118, 7118, 7119, 7119, 7119,
                                                                   7119, 7120, 7120, 7120, 7120, 7121, 7121, 7121, 7121, 7122, 7122,
                                                                   7122, 7122, 7123, 7123, 7123, 7123, 7124, 7124, 7124, 7124, 7125,
                                                                   7125, 7125, 7125, 7126, 7126, 7126, 7126, 7127, 7127, 7128, 7128,
                                                                   7129, 7129, 7130, 7130, 7131, 7131, 7132, 7132, 7133, 7133, 7134,
                                                                   7134, 7135, 7135, 7136, 7136, 7137, 7137, 7137, 7137, 7138, 7138,
                                                                   7139, 7139, 7139, 7139, 7140, 7140, 7141, 7141, 7141, 7141, 7142,
                                                                   7142, 7143, 7143, 7143, 7143, 7144, 7144, 7145, 7145, 7146, 7146,
                                                                   7146, 7146, 7147, 7147, 7147, 7147, 7148, 7148, 7148, 7148, 7149])
    elif current_station == 'DENHDR':
        expected_ts_ext_prediction_main_HWLWno_HWLWno = np.array([7116, 7116, 7117, 7117, 7118, 7118, 7119, 7119, 7120, 7120, 7121,
                                                                  7121, 7122, 7122, 7123, 7123, 7124, 7124, 7125, 7125, 7126, 7126,
                                                                  7127, 7127, 7128, 7128, 7129, 7129, 7130, 7130, 7131, 7131, 7132,
                                                                  7132, 7133, 7133, 7134, 7134, 7135, 7135, 7136, 7136, 7137, 7137,
                                                                  7138, 7138, 7139, 7139, 7140, 7140, 7141, 7141, 7142, 7142, 7143,
                                                                  7143, 7144, 7144, 7145, 7145, 7146, 7146, 7147, 7147, 7148, 7148])
        expected_ts_ext_prediction_main_HWLWcode = np.array([1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2,
                                                             1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2,
                                                             1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2])
        """
        expected_ts_ext_prediction_all_HWLWcode = np.array([11, 22,  1,  2, 11, 22,  1,  2, 11, 22,  1,  2, 11, 22,  1,  2, 11,
                                                            22,  1,  2, 11, 22,  1,  2, 11, 22,  1,  2, 11, 22,  1,  2,  1, 22,
                                                            11,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,
                                                             2,  1,  2,  1,  2,  1,  2,  1,  2,  1, 22, 11,  2,  1,  2, 11, 22,
                                                             1,  2,  1,  2, 11, 22,  1,  2,  1,  2,  1, 22, 11,  2,  1,  2,  1,
                                                             2,  1,  2,  1,  2,  1,  2, 11])
        """
        expected_ts_ext_prediction_clean_HWLWno_HWLWcode = np.array([1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2,
                                                                     1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2,
                                                                     1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2])
        expected_ts_ext_prediction_clean_HWLWno_HWLWno = np.array([7116, 7116, 7117, 7117, 7118, 7118, 7119, 7119, 7120, 7120, 7121,
                                                                   7121, 7122, 7122, 7123, 7123, 7124, 7124, 7125, 7125, 7126, 7126,
                                                                   7127, 7127, 7128, 7128, 7129, 7129, 7130, 7130, 7131, 7131, 7132,
                                                                   7132, 7133, 7133, 7134, 7134, 7135, 7135, 7136, 7136, 7137, 7137,
                                                                   7138, 7138, 7139, 7139, 7140, 7140, 7141, 7141, 7142, 7142, 7143,
                                                                   7143, 7144, 7144, 7145, 7145, 7146, 7146, 7147, 7147, 7148, 7148])
    
    assert np.allclose(ts_ext_prediction_main_HWLWno['HWLWno'].values, expected_ts_ext_prediction_main_HWLWno_HWLWno)
    assert np.allclose(ts_ext_prediction_main['HWLWcode'].values, expected_ts_ext_prediction_main_HWLWcode)
    assert np.allclose(ts_ext_prediction_clean_tomain['HWLWcode'].values, expected_ts_ext_prediction_main_HWLWcode)
    assert np.allclose(ts_ext_prediction_clean_HWLWno['HWLWcode'].values, expected_ts_ext_prediction_clean_HWLWno_HWLWcode)
    assert np.allclose(ts_ext_prediction_clean_HWLWno['HWLWno'].values, expected_ts_ext_prediction_clean_HWLWno_HWLWno)


@pytest.mark.systemtest
def test_hwlw_numbering_aggers():
    """
    This is probably already tested with other hwlw tests
    """
    
    times_pred = slice(dt.datetime(2010,1,31,3),dt.datetime(2010,2,17,12), "1min") #longer period with alternating aggers and no aggers, also eerste HW wordt als lokaal ipv primair HW gezien, also extra agger outside of 1stLW/agger/2ndLW sequence
    #HOEKVHLD has alternating aggers, DENHDR has double HW's
    expect_345_len = {'HOEKVHLD':99,'DENHDR':66}
    expect_345_code_sum = {'HOEKVHLD':267,'DENHDR':99}
    expect_345_nos_sum = {'HOEKVHLD':706061,'DENHDR':470712}
    
    for current_station in expect_345_len.keys():
    
        file_data_comp0 = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
        comp_merged = hatyan.read_components(filename=file_data_comp0)
        
        ts_prediction_HWLWno = hatyan.prediction(comp=comp_merged, times=times_pred)
        ts_ext_prediction_345 = hatyan.calc_HWLW(ts=ts_prediction_HWLWno, calc_HWLW345=True)
        ts_ext_prediction_345_HWLWno = hatyan.calc_HWLWnumbering(ts_ext=ts_ext_prediction_345, station=current_station)
        
        assert len(ts_ext_prediction_345_HWLWno) == expect_345_len[current_station]
        assert ts_ext_prediction_345_HWLWno["HWLWcode"].sum() == expect_345_code_sum[current_station]
        assert ts_ext_prediction_345_HWLWno["HWLWno"].sum() == expect_345_nos_sum[current_station]
        
        fig, (ax1,ax2) = hatyan.plot_timeseries(ts=ts_prediction_HWLWno, ts_ext=ts_ext_prediction_345)
        for irow, pdrow in ts_ext_prediction_345_HWLWno.iterrows():
            ax1.text(pdrow.name,pdrow['values'],pdrow['HWLWno'].astype(int))
        ax1.set_ylim(-1.2,1.7)
        hatyan.close()
        

@pytest.mark.systemtest
def test_19Ycomp4Ydia():
    # 1. define test data
    nodalfactors = True
    xfac = True
    const_list = hatyan.get_const_list_hatyan('year')
    current_station = 'VLISSGN'
    
    #comp0
    file_data_comp0 = os.path.join(dir_testdata,f'{current_station}_obs?.txt')
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    comp_frommeasurements_avg_group0 = hatyan.analysis(ts=ts_measurements_group0, const_list=const_list, nodalfactors=nodalfactors, xfac=xfac, fu_alltimes=False, analysis_perperiod='Y')
    comp_frommeasurements_avg_group0.station = current_station
    
    #comp1
    file_data_comp1 = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
    comp_fromfile_group1 = hatyan.read_components(filename=file_data_comp1)
    
    #merge component groups (SA/SM from 19Y, rest from 4Y)
    COMP_merged = hatyan.merge_componentgroups(comp_main=comp_frommeasurements_avg_group0, comp_sec=comp_fromfile_group1.loc[['SA','SM']])
    
    #prediction and validation
    times_pred = slice(dt.datetime(2019,1,1),dt.datetime(2019,1,1,12), "10min")
    ts_prediction = hatyan.prediction(COMP_merged, times=times_pred)
    ts_prediction_values = ts_prediction['values'].values
    
    # 2. define initial expectations (VLISSGN)
    expected_ts_prediction_data_pd_values = np.array([1.00071654,  0.88827903,  0.76819314,  0.64141558,  0.5093034 ,
                                                      0.37343441,  0.23542045,  0.09675753, -0.04125568, -0.17751967,
                                                     -0.31106673, -0.44095367, -0.56615373, -0.68548689, -0.79761632,
                                                     -0.9011202 , -0.99462701, -1.07698355, -1.14741452, -1.2056322 ,
                                                     -1.25186587, -1.28679942, -1.31142807, -1.32686565, -1.33414624,
                                                     -1.33406597, -1.3271011 , -1.31341934, -1.29297812, -1.2656809 ,
                                                     -1.23154773, -1.19085167, -1.14418076, -1.09240329, -1.03653868,
                                                     -0.97756028, -0.91617494, -0.85263162, -0.78660665, -0.71719668,
                                                     -0.64302677, -0.56245529, -0.47383656, -0.37579035, -0.26742836,
                                                     -0.14849976, -0.01943851,  0.11868132,  0.26425605,  0.41529083,
                                                      0.56952725,  0.72453204,  0.8777369 ,  1.02644009,  1.16779776,
                                                      1.29884123,  1.4165532 ,  1.51802108,  1.60066312,  1.6624979 ,
                                                      1.70240777,  1.720337  ,  1.71736989,  1.69565303,  1.65815518,
                                                      1.60829263,  1.54947792,  1.48467006,  1.41600766,  1.3445927 ,
                                                      1.27046406,  1.19276203,  1.11004681])

    # 4. Vefiry final expectations
    assert (np.abs(ts_prediction_values - expected_ts_prediction_data_pd_values) < 10E-9).all()


@pytest.mark.systemtest
def test_19Ycomp4Ydia_compsplitsing():
    # 1. define test data
    nodalfactors = True
    xfac = False
    const_list = hatyan.get_const_list_hatyan('month')
    current_station = 'D15'
    file_data_comp0 = os.path.join(dir_testdata,'%s_obs1.txt'%(current_station))
    cs_comps = pd.DataFrame({'CS_comps_from':['K1','N2','2MN2','S2','S2'],
                             'CS_comps_derive':['P1','NU2','LABDA2','K2','T2'],
                             'CS_ampfacs':[0.33,0.22,0.48,0.29,0.05],
                             'CS_degincrs':[-11,-24,174,1,-24]})
    file_data_comp1 = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
    times_pred = slice(dt.datetime(2019,1,1),dt.datetime(2019,1,1,12), "10min")
    
    # 3. run test
    #component groups
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    comp_frommeasurements_avg_group0 = hatyan.analysis(ts=ts_measurements_group0, nodalfactors=nodalfactors, xfac=xfac, 
                                                       const_list=const_list, analysis_perperiod=False, fu_alltimes=False, 
                                                       cs_comps=cs_comps)
    comp_fromfile_group1 = hatyan.read_components(filename=file_data_comp1)
    
    #merge component groups (SA/SM from 19Y, rest from 4Y)
    COMP_merged = hatyan.merge_componentgroups(comp_main=comp_frommeasurements_avg_group0, comp_sec=comp_fromfile_group1.loc[['SA','SM']])
    
    #prediction and validation
    ts_prediction = hatyan.prediction(COMP_merged, times=times_pred)
    ts_prediction_values = ts_prediction['values'].values
    
    # 2. define initial expectations (D15)
    expected_ts_prediction_data_pd_values = np.array([0.21208738,  0.26853271,  0.32342168,  0.37563977,  0.4242852 ,
                                                      0.46874611,  0.50873038,  0.54424431,  0.57552477,  0.60293749,
                                                      0.6268598 ,  0.64756857,  0.66515311,  0.67946822,  0.69013541,
                                                      0.69659205,  0.6981796 ,  0.69425554,  0.68430933,  0.66806236,
                                                      0.64553448,  0.61706564,  0.58328909,  0.54506078,  0.50335705,
                                                      0.45915817,  0.41333717,  0.36657245,  0.31929793,  0.27169799,
                                                      0.22374628,  0.17527983,  0.12609415,  0.0760412 ,  0.02511229,
                                                     -0.0265089 , -0.07843269, -0.1300906 , -0.18079823, -0.22984136,
                                                     -0.27656929, -0.32047779, -0.36126658, -0.39886054, -0.43339088,
                                                     -0.46513999, -0.49446005, -0.52168068, -0.54702266, -0.57053377,
                                                     -0.59205869, -0.61124872, -0.62760946, -0.64057802, -0.64961524,
                                                     -0.65429615, -0.65438144, -0.64985669, -0.64093152, -0.62799831,
                                                     -0.61155776, -0.59212467, -0.57013128, -0.54584644, -0.5193262 ,
                                                     -0.49040627, -0.45873912, -0.42387076, -0.38534492, -0.3428169 ,
                                                     -0.29615745, -0.24552815, -0.19141437])
    
    # 4. Vefiry final expectations
    assert (np.abs(ts_prediction_values - expected_ts_prediction_data_pd_values) < 10E-9).all()


@pytest.mark.systemtest
def test_allfromdia():
    # 1. define test data
    nodalfactors = True
    xfac = True
    const_list = hatyan.get_const_list_hatyan('year')
    current_station = 'VLISSGN'
    file_data_comp0 = [os.path.join(dir_testdata,'%s_obs%i.txt'%(current_station, file_id)) for file_id in [1,2,3,4]]
    file_data_comp1 = os.path.join(dir_testdata,'%s_obs19.txt'%(current_station))
    times_pred = slice(dt.datetime(2019,1,1),dt.datetime(2019,1,1,12), "10min")

    #component groups
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    comp_frommeasurements_avg_group0 = hatyan.analysis(ts=ts_measurements_group0, nodalfactors=nodalfactors, xfac=xfac, const_list=const_list, analysis_perperiod='Y', fu_alltimes=False)
    ts_measurements_group1 = hatyan.read_dia(filename=file_data_comp1, station=current_station)
    comp_frommeasurements_avg_group1 = hatyan.analysis(ts=ts_measurements_group1, nodalfactors=nodalfactors, xfac=xfac, const_list=const_list, analysis_perperiod=False, fu_alltimes=False)
    
    #merge component groups (SA/SM from 19Y, rest from 4Y)
    COMP_merged = hatyan.merge_componentgroups(comp_main=comp_frommeasurements_avg_group0, comp_sec=comp_frommeasurements_avg_group1.loc[['SA','SM']])
    
    #prediction and validation
    ts_prediction = hatyan.prediction(COMP_merged, times=times_pred)
    ts_prediction_values = ts_prediction['values'].values
    
    # 2. define initial expectations
    """expected_ts_prediction_data_pd_values_CADZD = np.array([  0.69689822,  0.58122683,  0.46063218,  0.3366989 ,  0.21084484,
                                                        0.08416485, -0.04259883, -0.16895802, -0.29448491, -0.41858223,
                                                       -0.54028668, -0.65816757, -0.77035852, -0.87472565, -0.96913982,
                                                       -1.05179199, -1.12147718, -1.17777783, -1.22109997, -1.25255039,
                                                       -1.27368164, -1.28616356, -1.29145829, -1.29057458, -1.28395729,
                                                       -1.27153413, -1.25290186, -1.22759891, -1.19538913, -1.15647792,
                                                       -1.11159891, -1.06194149, -1.00893047, -0.95390715, -0.89778804,
                                                       -0.84078547, -0.78226211, -0.72076111, -0.65421307, -0.5802802 ,
                                                       -0.49676665, -0.40200983, -0.29517405, -0.17639266, -0.04674223,
                                                        0.09192805,  0.23725639,  0.38664551,  0.53748768,  0.68728945,
                                                        0.83368193,  0.97433907,  1.10685673,  1.22865962,  1.33699851,
                                                        1.42907551,  1.50229865,  1.55462575,  1.58492517,  1.59326408,
                                                        1.58104106,  1.55090766,  1.50646777,  1.45179364,  1.39084215,
                                                        1.32688186,  1.26204443,  1.1970914 ,  1.13144293,  1.06345986,
                                                        0.99091634,  0.91155995,  0.82363968])
    """
    expected_ts_prediction_data_pd_values = np.array([1.00064431,  0.8882068 ,  0.76812092,  0.64134336,  0.50923119,
                                                      0.3733622 ,  0.23534825,  0.09668533, -0.04132787, -0.17759185,
                                                     -0.31113891, -0.44102584, -0.5662259 , -0.68555906, -0.79768847,
                                                     -0.90119235, -0.99469915, -1.07705569, -1.14748665, -1.20570432,
                                                     -1.25193799, -1.28687153, -1.31150017, -1.32693775, -1.33421834,
                                                     -1.33413806, -1.32717318, -1.31349142, -1.2930502 , -1.26575297,
                                                     -1.23161979, -1.19092373, -1.14425281, -1.09247534, -1.03661071,
                                                     -0.97763231, -0.91624696, -0.85270364, -0.78667866, -0.71726869,
                                                     -0.64309876, -0.56252728, -0.47390854, -0.37586232, -0.26750033,
                                                     -0.14857172, -0.01951046,  0.11860937,  0.26418411,  0.4152189 ,
                                                      0.56945533,  0.72446013,  0.87766499,  1.02636819,  1.16772586,
                                                      1.29876934,  1.41648132,  1.51794921,  1.60059126,  1.66242605,
                                                      1.70233593,  1.72026516,  1.71729806,  1.6955812 ,  1.65808337,
                                                      1.60822082,  1.54940611,  1.48459827,  1.41593587,  1.34452092,
                                                      1.27039229,  1.19269026,  1.10997505])
    
    # 4. Vefiry final expectations
    assert (np.abs(ts_prediction_values - expected_ts_prediction_data_pd_values) < 10E-9).all()


@pytest.mark.systemtest
def test_allfromdia_2008xfac0():
    # 1. define test data
    nodalfactors = True
    xfac = False
    const_list = hatyan.get_const_list_hatyan('year')
    current_station = 'DOVR'
    file_data_comp0 = os.path.join(dir_testdata,'%s_obs1.txt'%(current_station))
    times_pred = slice(dt.datetime(2019,1,1),dt.datetime(2019,1,1,12), "10min")
    
    #component groups
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    comp_frommeasurements_avg_group0 = hatyan.analysis(ts=ts_measurements_group0, nodalfactors=nodalfactors, xfac=xfac, const_list=const_list, analysis_perperiod='Y', fu_alltimes=False)
    
    #prediction and validation
    ts_prediction = hatyan.prediction(comp=comp_frommeasurements_avg_group0, times=times_pred)
    ts_prediction_values = ts_prediction['values'].values
    
    # 2. define initial expectations
    expected_ts_prediction_data_pd_values = np.array([-0.60440198, -0.72142681, -0.83581976, -0.94743404, -1.05609766,
                                                      -1.1615773 , -1.2635405 , -1.36152022, -1.45488748, -1.54283832,
                                                      -1.62440113, -1.69846786, -1.76384908, -1.81934804, -1.86384401,
                                                      -1.89637159, -1.91618096, -1.92276574, -1.91584961, -1.89533042,
                                                      -1.86118924, -1.81338063, -1.75172686, -1.67584147, -1.58510561,
                                                      -1.4787132 , -1.35578995, -1.21557736, -1.05765978, -0.88220159,
                                                      -0.69015614, -0.48340872, -0.26482336, -0.03817622,  0.19202489,
                                                       0.42081806,  0.64312991,  0.8541411 ,  1.04962903,  1.22624559,
                                                       1.38169767,  1.514813  ,  1.62549027,  1.7145481 ,  1.78350005,
                                                       1.83429032,  1.86902593,  1.88973797,  1.89819605,  1.89578968,
                                                       1.88347976,  1.86181361,  1.83099019,  1.79095846,  1.74153131,
                                                       1.68249919,  1.61373091,  1.53525334,  1.44730507,  1.35036266,
                                                       1.24514015,  1.13256443,  1.01373016,  0.8898392 ,  0.76213036,
                                                       0.63180616,  0.49996392,  0.36753791,  0.23525891,  0.10363497,
                                                      -0.02704478, -0.15668462, -0.28533886])
    
    # 4. Vefiry final expectations
    assert (np.abs(ts_prediction_values - expected_ts_prediction_data_pd_values) < 10E-9).all()





# -*- coding: utf-8 -*-
"""
Created on Wed Aug  9 12:14:02 2023

@author: veenstra
"""

import os
import pytest
import pandas as pd
import datetime as dt
import numpy as np
from netCDF4 import Dataset, num2date
import hatyan
from hatyan.metadata import metadata_from_obj, metadata_compare

dir_tests = os.path.dirname(__file__) #F9 doesnt work, only F5 (F5 also only method to reload external definition scripts)
dir_testdata = os.path.join(dir_tests,'data_unitsystemtests')


@pytest.mark.unittest
def test_get_diaxycoords():
    file_pred = os.path.join(dir_testdata,"VLISSGN_pre.txt")
    stat_x, stat_y = hatyan.get_diaxycoords(filename=file_pred, crs=28992)
    assert np.isclose(stat_x, 30480.0)
    assert np.isclose(stat_y, 385220.0)
    
    file_pred = os.path.join(dir_testdata,"hoek_har.dia")
    stat_x, stat_y = hatyan.get_diaxycoords(filename=file_pred, crs=28992)
    assert np.allclose(stat_x, [67930., 67930., 67930., 49862.])
    assert np.allclose(stat_y, [444000., 444000., 444000., 431612.])


@pytest.mark.unittest
def test_readwrite_tsdia_rounding():
    file_pred = os.path.join(dir_testdata, "VLISSGN_pre.txt")
    file_new = 'temp_dia.txt'
    ts_pred = hatyan.read_dia(file_pred)
    ts_pred.pop('status')
    for diff in [0,0.004, -0.004]:
        ts_pred_rounddiff = hatyan.read_dia(file_pred)
        ts_pred_rounddiff['values'] = ts_pred['values'] + diff
        hatyan.write_dia(ts=ts_pred_rounddiff, filename=file_new)
        ts_new = hatyan.read_dia(file_new)
        ts_new.pop('status')
        assert np.allclose(ts_pred, ts_new)
    os.remove(file_new)


@pytest.mark.unittest
def test_readwrite_tsdia_ext_rounding():
    file_pred = os.path.join(dir_testdata, "VLISSGN_ext.txt")
    file_new = 'temp_dia_ext.txt'
    ts_pred = hatyan.read_dia(file_pred)
    ts_pred.pop('status')
    for diff in [0, 0.004, -0.004]:
        ts_pred_rounddiff = hatyan.read_dia(file_pred)
        ts_pred_rounddiff['values'] = ts_pred['values'] + diff
        hatyan.write_dia(ts=ts_pred_rounddiff, filename=file_new)
        ts_new = hatyan.read_dia(file_new)
        ts_new.pop('status')        
        assert np.allclose(ts_pred, ts_new)
    os.remove(file_new)


@pytest.mark.unittest
def test_readts_dia_multifile():
    file_data_comp0 = [os.path.join(dir_testdata,f'VLISSGN_obs{file_id}.txt') for file_id in [1,2,3,4]]
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0, station='VLISSGN')
    
    assert len(ts_measurements_group0) == 35064
    assert np.isclose(ts_measurements_group0['values'].iloc[0], -1.24)
    assert np.isclose(ts_measurements_group0['values'].iloc[-1], -1.5)
    assert (ts_measurements_group0['qualitycode'] != 0).sum() == 82
    assert list(np.unique(ts_measurements_group0['qualitycode'])) == [ 0, 25]


@pytest.mark.unittest
def test_readts_dia_multifile_multiblock():
    """
    this ordering of files raised `ValueError: Invalid values in block_ids list ([0, 1, 2]), possible are [0] (all integers)`
    since block_ids was overwritten in the file loop before. After a fix we still get an error, but a different one.
    """
    file_dia1 = os.path.join(dir_testdata,"hoek_har.dia")
    file_dia2 = os.path.join(dir_testdata,"diawia_HOEKVHLD_astro_extremen.dia")
    with pytest.raises(ValueError) as e:
        hatyan.read_dia([file_dia1,file_dia2], block_ids="allstation", station="HOEKVHLD")
    assert 'metadata for two datasets is not equal, cannot be merged:' in str(e.value)


@pytest.mark.unittest
def test_readts_dia_multiblock():
    
    file1 = os.path.join(dir_testdata,'hoek_har.dia')
    ts_measurements_group0_ext0 = hatyan.read_dia(filename=file1, station='HOEKVHLD', block_ids=0)
    ts_measurements_group0_ext1 = hatyan.read_dia(filename=file1, station='HOEKVHLD', block_ids=1)
    ts_measurements_group0_ext2 = hatyan.read_dia(filename=file1, station='HOEKVHLD', block_ids=2)
    ts_measurements_group0_ext012 = hatyan.read_dia(filename=file1, station='HOEKVHLD', block_ids=[0,1,2])
    ts_measurements_group0_extall = hatyan.read_dia(filename=file1, station='HOEKVHLD', block_ids='allstation')
    
    assert len(ts_measurements_group0_ext0) == 3977
    assert len(ts_measurements_group0_ext1) == 9913
    assert len(ts_measurements_group0_ext2) == 9403
    assert len(ts_measurements_group0_ext012) == 23293
    assert len(ts_measurements_group0_extall) == 23293


@pytest.mark.unittest
def test_readts_dia_multiblock_varyingtimestep(tmp_path):
    """
    implemented while fixing https://github.com/Deltares/hatyan/issues/313
    """
    file1 = os.path.join(dir_testdata, "diawia_HOEKVHLD_astro_tijdreeks.dia") # 10min interval
    file2 = os.path.join(dir_testdata, "HOEKVHLD_obs19.txt") # 60min interval
    file_dia = os.path.join(tmp_path, "HOEKVHLD_merged.dia")
    with open(file_dia, "w") as f:
        with open(file1, "r") as f1:
            contents = f1.read()
            contents_new = contents.replace("WATHTBRKD","WATHTE")
            f.write(contents_new)
        with open(file2, "r") as f2:
            contents = f2.readlines()
            contents_nohead = contents[1:]
            contents_new = ''.join(contents_nohead)
            f.write(contents_new)
    ts = hatyan.read_dia(file_dia, block_ids="allstation", station="HOEKVHLD")
    assert len(ts) == 219264
    # check whether min/max attributes are correct since they are derived from the dataset
    assert ts.index.min() == pd.Timestamp('1976-01-01 00:00:00 +01:00')
    assert ts.index.max() == pd.Timestamp('2020-12-31 23:50:00 +01:00')


@pytest.mark.unittest
def test_read_dia_multiblock_toolittle_arguments():
    file_dia = os.path.join(dir_testdata,'hoek_har.dia')
    
    with pytest.raises(ValueError) as e:
        hatyan.read_dia(filename=file_dia)
    assert 'If block_ids=None or block_ids="allstation", station argument should be provided.' in str(e.value)
    
    with pytest.raises(ValueError) as e:
        hatyan.read_dia(filename=file_dia, station='HOEKVHLD')
    assert "More than one data block with requested station (HOEKVHLD) present in dia file." in str(e.value)
    
    with pytest.raises(ValueError) as e:
        hatyan.read_dia(filename=file_dia, station='VLISSGN')
    assert "No data block with requested station (VLISSGN) present in dia file." in str(e.value)


@pytest.mark.unittest
def test_readwrite_noos():

    file_data_comp0 = os.path.join(dir_testdata,'VLISSGN_waterlevel_20180101_20180401.noos')
    
    #component groups
    filename_out = "noos_test.txt"
    ts_noos1 = hatyan.read_noos(filename=file_data_comp0)
    hatyan.write_noos(ts_noos1, filename=filename_out)
    ts_noos2 = hatyan.read_noos(filename=filename_out)
    
    assert (ts_noos1 == ts_noos2).all().all()
    
    # compare metadata, raises ValueError if not equal
    meta1 = metadata_from_obj(ts_noos1)
    meta2 = metadata_from_obj(ts_noos2)
    metadata_compare([meta1, meta2])
    os.remove(filename_out)


@pytest.mark.unittest
def test_readts_noos_resamplecrop():

    file_data_comp0 = os.path.join(dir_testdata,'VLISSGN_waterlevel_20180101_20180401.noos')
    times_ext_comp0 = slice(dt.datetime(2018,1,1),dt.datetime(2018,4,1))

    #component groups
    ts_measurements_group0 = hatyan.read_noos(filename=file_data_comp0)
    ts_measurements_group0_res = hatyan.resample_timeseries(ts_measurements_group0, timestep_min=10)
    ts_measurements_group0_rescrop = hatyan.crop_timeseries(ts_measurements_group0_res, times=times_ext_comp0)
    
    assert len(ts_measurements_group0) == 12752
    assert len(ts_measurements_group0_res) == 12961
    assert len(ts_measurements_group0_rescrop) == 12961
    assert ts_measurements_group0_rescrop.index[0] == pd.Timestamp('2018-01-01')
    assert ts_measurements_group0_rescrop.index[-1] == pd.Timestamp('2018-04-01')
    assert np.isclose(ts_measurements_group0_rescrop['values'].iloc[0], 2.5)
    assert np.isclose(ts_measurements_group0_rescrop['values'].iloc[-1], 1.05)


@pytest.mark.unittest
def test_readts_dia_equidistant_singlefile_hasfreq():
    """
    When reading asingle equidistant diafile,
    there should be a freq attribute that is not None
    """
    file_ts = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    
    ts_pd = hatyan.read_dia(filename=file_ts)
    
    # check ts length (all four files are added)
    assert len(ts_pd) == 8760
    
    # assert on freq attribute
    assert hasattr(ts_pd.index,'freq')
    assert isinstance(ts_pd.index.freq,pd.offsets.Minute)
    assert ts_pd.index.freq is not None
    assert np.isclose(ts_pd.index.freq.nanos/1e9, 3600)


@pytest.mark.unittest
def test_pandas_concat_hasfreq():
    """
    freq is None in case of index with non-constant freq (eg 2020+2022 or if one timestep is skipped/duplicated)
    freq is pd.offsets.Minute if index is equidistant
    """

    def df_index(year):
        pd_year = pd.date_range(f'{year}-01-01',f'{year}-12-31 23:50', freq='10min')
        df_year = pd.DataFrame(index=pd_year)
        return df_year
    
    df_2020 = df_index(2020)
    df_2021 = df_index(2021)
    df_2022 = df_index(2022)
    
    ts_pd = pd.concat([df_2020,df_2021])
    ts_pd_nonequi = pd.concat([df_2020,df_2022])
    
    # assert on freq attribute
    assert hasattr(ts_pd.index,'freq')
    assert isinstance(ts_pd.index.freq,pd.offsets.Minute)
    assert ts_pd.index.freq is not None
    assert np.isclose(ts_pd.index.freq.nanos/1e9, 600)

    # assert on freq attribute
    assert hasattr(ts_pd_nonequi.index,'freq')
    assert isinstance(ts_pd_nonequi.index.freq,type(None))
    assert ts_pd_nonequi.index.freq is None


@pytest.mark.unittest
def test_readts_dia_equidistant_multifile_hasfreq():
    """
    When reading multiple equidistant diafiles that combine into a continuous timeseries,
    there should be a freq attribute that is not None
    """
    file_ts = [os.path.join(dir_testdata,f'VLISSGN_obs{i}.txt') for i in [1,2,3,4]]
    
    ts_pd = hatyan.read_dia(filename=file_ts)
    
    # check ts length (all four files are added)
    assert len(ts_pd) == 35064
    
    # check index dtype
    assert ts_pd.index.dtype == "datetime64[ns, pytz.FixedOffset(60)]"
    
    # checks for freq attribute
    assert hasattr(ts_pd.index,'freq')
    assert isinstance(ts_pd.index.freq,pd.offsets.Minute)
    assert ts_pd.index.freq is not None
    assert np.isclose(ts_pd.index.freq.nanos/1e9, 3600)


@pytest.mark.unittest
def test_readts_dia_equidistant_multifile_glob_hasfreq():
    """
    When providing a file pattern for reading multiple equidistant diafiles,
    glob is supported, this test checks if it results in the correct data
    """
    file_ts = os.path.join(dir_testdata,'VLISSGN_obs?.txt')
    
    ts_pd = hatyan.read_dia(filename=file_ts)
    
    assert len(ts_pd) == 35064

    # check index dtype
    assert ts_pd.index.dtype == "datetime64[ns, pytz.FixedOffset(60)]"
    
    # checks for freq attribute
    assert hasattr(ts_pd.index,'freq')
    assert isinstance(ts_pd.index.freq,pd.offsets.Minute)
    assert ts_pd.index.freq is not None
    assert np.isclose(ts_pd.index.freq.nanos/1e9, 3600)


@pytest.mark.unittest
def test_readwrite_diawia():
    
    current_station = 'HOEKVHLD'
    
    file_dia_wl = os.path.join(dir_testdata,'diawia_%s_astro_tijdreeks.dia'%(current_station))
    file_dia_ext = os.path.join(dir_testdata,'diawia_%s_astro_extremen.dia'%(current_station))
    
    for file_dia in [file_dia_wl,file_dia_ext]:
        file_wia = file_dia.replace('.dia','.wia')
        file_dia_out = file_dia.replace('.dia','_out.dia')
        file_wia_out = file_dia.replace('.dia','_out.wia')
        ts_dia = hatyan.read_dia(filename=file_dia, station=current_station)
        ts_wia = hatyan.read_dia(filename=file_wia, station=current_station)
        assert (ts_dia==ts_wia).all().all() #check if wia and dia input is equal
        
        #write to files
        hatyan.write_dia(ts=ts_dia, filename=file_dia_out)
        hatyan.write_dia(ts=ts_wia, filename=file_wia_out, headerformat='wia')
        
        #read from new files
        ts_dia_new = hatyan.read_dia(filename=file_dia_out, station=current_station)
        ts_wia_new = hatyan.read_dia(filename=file_wia_out, station=current_station)
        ts_dia.pop('status')
        ts_wia.pop('status')
        ts_dia_new.pop('status')
        ts_wia_new.pop('status')
        assert np.allclose(ts_dia, ts_dia_new) #check if wia and dia input is equal
        assert np.allclose(ts_wia, ts_wia_new) #check if wia and dia input is equal
        
        #meta
        meta_dia = metadata_from_obj(ts_dia)
        meta_wia = metadata_from_obj(ts_wia)
        meta_dia_new = metadata_from_obj(ts_dia_new)
        meta_wia_new = metadata_from_obj(ts_wia_new)
        assert (meta_dia == meta_wia == meta_dia_new == meta_wia_new)
        
        #remove files
        os.remove(file_dia_out)
        os.remove(file_wia_out)


@pytest.mark.unittest
def test_crop_timeseries():
    
    current_station = 'VLISSGN'
    times_ext = slice("2019-01-01 00:00:00 +00:00", "2019-06-01 00:00:00 +00:00")
    
    file_pred = os.path.join(dir_testdata,f'{current_station}_pre.txt')
    ts_prediction = hatyan.read_dia(filename=file_pred, station=current_station)
    ts_prediction_cropped = hatyan.crop_timeseries(ts_prediction, times=times_ext)
    
    assert len(ts_prediction_cropped) == 21745
    assert ts_prediction_cropped.index[0] == pd.Timestamp(times_ext.start)
    assert ts_prediction_cropped.index[-1] == pd.Timestamp(times_ext.stop)
    
    pred_meta = metadata_from_obj(ts_prediction)
    pred_cropped_meta = metadata_from_obj(ts_prediction_cropped)
    metadata_compare([pred_meta,pred_cropped_meta])


@pytest.mark.unittest
def test_resample_timeseries():
    
    current_station = 'VLISSGN'
    timestep_min = 120
    
    file_pred = os.path.join(dir_testdata,f'{current_station}_pre.txt')
    ts_prediction = hatyan.read_dia(filename=file_pred, station=current_station)
    ts_prediction_res = hatyan.resample_timeseries(ts_prediction, timestep_min=timestep_min)
    
    assert len(ts_prediction_res) == 4380
    assert ts_prediction_res.index[0] == pd.Timestamp(ts_prediction.index[0])
    assert ts_prediction_res.index[-1] == pd.Timestamp("2019-12-31 22:00 +01:00")
    
    pred_meta = metadata_from_obj(ts_prediction)
    pred_res_meta = metadata_from_obj(ts_prediction_res)
    metadata_compare([pred_meta,pred_res_meta])
    assert np.isclose(pd.Timedelta(ts_prediction_res.index.freq).total_seconds(), 60*timestep_min)


@pytest.mark.unittest
def test_write_tsdia_rounding():
    """
    rounding error occurred in older versions of hatyan2
    Therefore we test whether writing and reading a timeseries results in the same data (accuracy of 1cm)
    """
    
    current_station = 'HOEKVHLD'
    file_data_comp0 = os.path.join(dir_testdata,f'{current_station}_ana.txt')
    
    times_pred = slice(dt.datetime(2019,1,1),dt.datetime(2020,1,1), "10min")
    
    comp_merged = hatyan.read_components(filename=file_data_comp0)
    
    #prediction and validation
    ts_prediction = hatyan.prediction(comp=comp_merged, times=times_pred)
    
    #write to file
    fname_pred = 'prediction_%s_%s.dia'%(times_pred.step,current_station)
    hatyan.write_dia(ts=ts_prediction, filename=fname_pred)
    
    #read from file
    ts_prediction_fromfile = hatyan.read_dia(filename=fname_pred, station=current_station)
    os.remove(fname_pred)
    
    # assert max differences
    ts_diff = ts_prediction_fromfile - ts_prediction
    assert (ts_diff['values'].abs()<=0.005).all()


@pytest.mark.unittest
def test_timeseries_fft():
    file_pred = os.path.join(dir_testdata, "VLISSGN_pre.txt")
    ts_pred = hatyan.read_dia(file_pred)
        
    hatyan_freqs_suggestions_schureman = hatyan.timeseries_fft(ts_pred, min_prominence=2000, plot_fft=True, source="schureman")
    hatyan_freqs_suggestions_foreman = hatyan.timeseries_fft(ts_pred, min_prominence=2000, plot_fft=True, source="foreman")
    const_list_schureman = hatyan_freqs_suggestions_schureman.index.unique().tolist()
    const_list_foreman = hatyan_freqs_suggestions_foreman.index.unique().tolist()
    
    assert const_list_schureman == ['O1', 'N2', 'M2', 'L2B', 'S2', 'K2', 'M4', 'M6']
    assert const_list_foreman ==   ['O1', 'N2', 'M2', 'L2', 'S2', 'K2', 'M4', 'M6']


@pytest.mark.unittest
def test_plot_timeseries():
    file_pred = os.path.join(dir_testdata, "VLISSGN_pre.txt")
    file_ext = os.path.join(dir_testdata, "VLISSGN_ext.txt")
    ts_pred = hatyan.read_dia(file_pred)
    ts_ext = hatyan.read_dia(file_ext)
    hatyan.plot_timeseries(ts=ts_pred, ts_validation=ts_pred, ts_ext=ts_ext, ts_ext_validation=ts_ext)


@pytest.mark.unittest
def test_plot_HWLW_validatestats():
    file_ext = os.path.join(dir_testdata, "VLISSGN_ext.txt")
    ts_ext = hatyan.read_dia(file_ext)
    ts_ext_nos = hatyan.calc_HWLWnumbering(ts_ext)

    hatyan.plot_HWLW_validatestats(ts_ext=ts_ext, ts_ext_validation=ts_ext)
    hatyan.plot_HWLW_validatestats(ts_ext=ts_ext_nos, ts_ext_validation=ts_ext_nos)
    
    # check if original dataframe was not altered
    assert "HWLWno" not in ts_ext.columns
    assert "HWLWno" in ts_ext_nos.columns


@pytest.mark.unittest
def test_nyquist_folding():
    const_list = hatyan.get_const_list_hatyan('year')
    
    # drop_list = ['S4','3M2S10','2SM6','4M2S12'] # overlappende frequenties na folding, en S4 is precies op Nyquist frequentie
    # for const in drop_list:
    #     const_list.remove(const)
    
    station = 'DENHDR'
    year = 1955
    url_dataraw = 'https://watersysteemdata.deltares.nl/thredds/fileServer/watersysteemdata/Wadden/ddl/raw/waterhoogte/'
    file_csv = url_dataraw+f'{station}_OW_WATHTE_NVT_NAP_{year}_ddl_wq.csv'
    
    data_pd = pd.read_csv(file_csv,sep=';',parse_dates=['tijdstip'])
    
    ts_meas_raw = pd.DataFrame({'values':data_pd['numeriekewaarde'].values/100},index=data_pd['tijdstip'].dt.tz_localize(None)) #read dataset and convert to DataFrame with correct format #TODO: tijdzone is MET (UTC+1), ook van de resulterende getijcomponenten. Is dat wenselijk?
    ts_meas_raw = ts_meas_raw.sort_index(axis='index') #sort dataset on times (not necessary but easy for debugging)
    
    # check if we get "Exception: there is a component on the Nyquist frequency (0.16666666666666666 [1/hr]), this not possible"
    # without doing the nyquist check, we get a MatrixConditionTooHigh error instead
    with pytest.raises(ValueError) as e:
        hatyan.analysis(ts=ts_meas_raw, const_list=const_list, nodalfactors=True, xfac=True, fu_alltimes=True)
    assert "nyquist" in str(e.value).lower()


@pytest.mark.unittest
def test_writenetcdf():
    
    current_station = 'VLISSGN'
    
    file_pred = os.path.join(dir_testdata,f'{current_station}_pre.txt')
    ts_prediction = hatyan.read_dia(filename=file_pred, station=current_station)
    ts_ext_prediction = hatyan.calc_HWLW(ts=ts_prediction)
    
    file_nc = 'prediction_10m_%s.nc'%(current_station)
    hatyan.write_netcdf(ts=ts_prediction, ts_ext=ts_ext_prediction, filename=file_nc)
    
    data_nc = Dataset(file_nc,'r')
    
    timevar = data_nc.variables['time']
    timevar_dt = num2date(timevar[:],units=timevar.units, only_use_cftime_datetimes=False, only_use_python_datetimes=True)
    
    #put netcdf file contents in pandas DataFrame for usage in hatyan
    ts_pd = pd.DataFrame({'values':data_nc.variables['waterlevel_astro'][:,0]}, index=timevar_dt)
    print(ts_pd)
    
    assert list(data_nc.dimensions.keys()) == ['stations', 'statname_len', 'time', 'analysis_time', 'time_HW', 'time_LW']
    assert list(data_nc.variables.keys()) == ['stations', 'analysis_time', 'time', 'waterlevel_astro', 'time_HW', 'waterlevel_astro_HW', 'time_LW', 'waterlevel_astro_LW']
    assert timevar_dt[0] == ts_prediction.index[0].tz_convert(None)
    assert timevar_dt[-1] == ts_prediction.index[-1].tz_convert(None)
    assert 'title' in data_nc.__dict__.keys()
    assert 'institution' in data_nc.__dict__.keys()
    assert 'source' in data_nc.__dict__.keys()
    assert timevar.units == 'minutes since 1900-01-01 00:00:00 +0100'

    data_nc.close()
    os.remove(file_nc)


@pytest.mark.unittest
def test_writenetcdf_nosidx():
    current_station = 'VLISSGN'
    
    file_pred = os.path.join(dir_testdata,f'{current_station}_pre.txt')
    ts_prediction = hatyan.read_dia(filename=file_pred, station=current_station)
    ts_ext_prediction = hatyan.calc_HWLW(ts=ts_prediction)
    ts_ext_prediction = hatyan.calc_HWLWnumbering(ts_ext=ts_ext_prediction)
    
    file_nc = 'prediction_10m_%s.nc'%(current_station)
    hatyan.write_netcdf(ts=ts_prediction, ts_ext=ts_ext_prediction, filename=file_nc, nosidx=True)
    
    data_nc = Dataset(file_nc,'r')
    
    timevar = data_nc.variables['time']
    timevar_dt = num2date(timevar[:],units=timevar.units, only_use_cftime_datetimes=False, only_use_python_datetimes=True)
    
    #put netcdf file contents in pandas DataFrame for usage in hatyan
    ts_pd = pd.DataFrame({'values':data_nc.variables['waterlevel_astro'][:,0]}, index=timevar_dt)
    print(ts_pd)
    
    assert list(data_nc.dimensions.keys()) == ['stations', 'statname_len', 'time', 'analysis_time', 'HWLWno']
    assert list(data_nc.variables.keys()) == ['stations', 'analysis_time', 'time', 'waterlevel_astro', 
                                              'HWLWno', 'times_astro_HW', 'waterlevel_astro_HW', 'times_astro_LW', 'waterlevel_astro_LW']
    assert timevar_dt[0] == ts_prediction.index[0].tz_convert(None)
    assert timevar_dt[-1] == ts_prediction.index[-1].tz_convert(None)
    assert 'title' in data_nc.__dict__.keys()
    assert 'institution' in data_nc.__dict__.keys()
    assert 'source' in data_nc.__dict__.keys()
    assert timevar.units == 'minutes since 1900-01-01 00:00:00 +0100'

    data_nc.close()
    os.remove(file_nc)


@pytest.mark.unittest
def test_calc_HWLW12345to12():
    file_ext = os.path.join(dir_testdata,'hoek_har.dia')
    
    df = hatyan.read_dia(file_ext, block_ids=0)
    df_12 = hatyan.calc_HWLW12345to12(df)
    
    assert len(df) == 3977
    assert len(df_12) == 2825


@pytest.mark.unittest
def test_calc_HWLW12345to12_include_last_lw():
    file_ext = os.path.join(dir_testdata,'hoek_har.dia')
    
    df = hatyan.read_dia(file_ext, block_ids=0)
    
    # get timeseries that ends with HWLWcode=2
    df_sel = df.iloc[:-1]
    df_12 = hatyan.calc_HWLW12345to12(df_sel)
    
    assert len(df) == 3977
    assert len(df_sel) == 3976
    assert df_sel["HWLWcode"].iloc[0] == 1
    assert df_sel["HWLWcode"].iloc[-1] == 2
    assert len(df_12) == 2824
    assert df_12["HWLWcode"].iloc[0] == 1
    assert df_12["HWLWcode"].iloc[-1] == 2


@pytest.mark.unittest
def test_calc_HWLW12345to12_include_first_lw():
    file_ext = os.path.join(dir_testdata,'hoek_har.dia')
    
    df = hatyan.read_dia(file_ext, block_ids=0)
    
    # get timeseries that starts with HWLWcode=2
    df_sel = df.iloc[9:]
    df_12 = hatyan.calc_HWLW12345to12(df_sel)
    
    assert len(df) == 3977
    assert len(df_sel) == 3968
    assert df_sel["HWLWcode"].iloc[0] == 2
    assert df_sel["HWLWcode"].iloc[-1] == 1
    assert len(df_12) == 2820
    assert df_12["HWLWcode"].iloc[0] == 2
    assert df_12["HWLWcode"].iloc[-1] == 1


@pytest.mark.unittest
def test_calc_HWLW12345to12_skip_missing_lw():
    file_ext = os.path.join(dir_testdata,'hoek_har.dia')
    
    df = hatyan.read_dia(file_ext, block_ids=0)
    
    # construct boolean to drop the first low waters (345 combination)
    # and the last low water (2)
    bool_drop = df["HWLWcode"] != 0
    bool_drop.iloc[1:4] = False
    bool_drop.iloc[-2:-1] = False
    
    df_sel = df.loc[bool_drop]
    df_12 = hatyan.calc_HWLW12345to12(df_sel)
    
    assert len(df) == 3977
    assert len(df_sel) == 3973
    assert df_sel["HWLWcode"].iloc[0:2].tolist() == [1,1]
    assert df_sel["HWLWcode"].iloc[-2:].tolist() == [1,1]
    assert len(df_12) == 2823
    assert df_12["HWLWcode"].iloc[0:2].tolist() == [1,1]
    assert df_12["HWLWcode"].iloc[-2:].tolist() == [1,1]


@pytest.mark.unittest
def test_calc_HWLW12345to12_already_12():
    file_ext = os.path.join(dir_testdata,'VLISSGN_ext.txt')
    
    df = hatyan.read_dia(file_ext)
    
    df_12 = hatyan.calc_HWLW12345to12(df)
    
    assert len(df) == 1411
    assert len(df_12) == 1411



# -*- coding: utf-8 -*-
"""
Created on Wed May  8 13:37:10 2024

@author: veenstra
"""

import pytest
import hatyan

list_deprecated_funcs = hatyan.deprecated.__all__

def test_deprecated_functions():
    for deprecated_func in list_deprecated_funcs:
        with pytest.raises(DeprecationWarning):
            exec(f"hatyan.{deprecated_func}()")



# -*- coding: utf-8 -*-

"""
Console script for hatyan.
    - ``hatyan --help``
"""
import sys
import logging
import click
import hatyan
import os
import shutil
import datetime as dt
import matplotlib
import matplotlib.pyplot as plt
import importlib

logger = logging.getLogger(__name__)


@click.command()
@click.argument(
    'filename',
    type=click.Path(exists=True),
)
@click.option(
    "-o", "--overwrite",
    is_flag=True,
    help="overwrite `dir_output` if it exists"
)
@click.option(
    "-i", "--interactive-plots",
    is_flag=True,
    help="show interactive plots at the end of the hatyan process"
)
@click.option(
    "-r", "--redirect-stdout",
    is_flag=True,
    help="redirecting stdout to dir_output/STDOUT.txt, "
    "warnings/errors are still printed to console"
)
@click.option(
    '-l',
    '--loglevel',
    type=str,
    help="set logging level to ERROR/WARNING/INFO/DEBUG, default is INFO",
)
@click.version_option(hatyan.__version__)
def cli(filename, overwrite, interactive_plots, redirect_stdout, loglevel):
    """
    Initializes hatyan by creating a `dir_output`, setting current and matplotlib
    savefig directory to `dir_output` and printing the initialisation header.
    Then runs the provided configfile (FILENAME).
    Wraps up hatyan by printing a de-initialisation footer with the script runtime.
    """
    
    # get file_config and start time
    file_config = os.path.abspath(filename)
    timer_start = dt.datetime.now()
    
    # create dir_output and chdir
    dir_output = os.path.splitext(os.path.basename(file_config))[0]
    if os.path.exists(dir_output) and not overwrite:
        raise FileExistsError(f"Directory '{dir_output}' already exists. Use '--overwrite' if you want to overwrite existing results.")
    os.makedirs(dir_output, exist_ok=True)
    os.chdir(dir_output)
    
    #copy configfile to cwd (=dir_output)
    shutil.copyfile(file_config, os.path.basename(file_config))
    
    #set the storage location of interactive plots
    matplotlib.rcParams["savefig.directory"] = dir_output
    
    #redirecting stdout, stderr is still printed to console
    if redirect_stdout:
        sys.stdout = open('STDOUT.txt', "w")
    
    # set logging level and stdout
    if loglevel is None:
        loglevel = "INFO"
    logging.basicConfig(level=loglevel, stream=sys.stdout)
    
    # initialization print
    logger.info("############### HATYAN INITALIZING ###############")
    logger.info("--------------------------------------------------")
    logger.info(f"hatyan-{hatyan.__version__}: RWS tidal analysis and prediction")
    logger.info(f"started at:  {timer_start.strftime('%Y-%m-%d %H:%M:%S')}")
    logger.info(f"file_config: {file_config}")
    logger.info(f"dir_output:  {dir_output}")
    logger.info("--------------------------------------------------")
    
    # run the configfile
    # exec from within cli somehow does not support oneline list generation with predefined variables
    # with open(file_config) as f:
    #     exec(f.read())
    # therefore we use importlib instead: https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
    spec = importlib.util.spec_from_file_location("arbitrary.name", file_config)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    
    # get stop time
    timer_stop = dt.datetime.now()
    timer_elapsed = (timer_stop - timer_start).total_seconds()/60
    
    # de-initialization print
    logger.info("--------------------------------------------------")
    logger.info(f"finished at: {timer_stop.strftime('%Y-%m-%d %H:%M:%S')}")
    logger.info(f"elapsed time: {timer_elapsed:.2f} minutes {timer_elapsed*60:.2f} seconds")
    logger.info("--------------------------------------------------")
    logger.info("################# HATYAN FINISHED ################")
    
    # show all created plots in case of interactive-plots
    if interactive_plots:
        logger.warning(f"close open plots to continue (mpl.backend='{matplotlib.get_backend()}')")
        plt.show()
    os.chdir("..")




# -*- coding: utf-8 -*-
"""
Created on Wed Mar 13 14:24:07 2024

@author: veenstra
"""

import pytest
import hatyan
import ddlpy
import datetime as dt
import numpy as np
import pandas.api.types as ptypes


@pytest.fixture(scope="session")
def locations():
    """return all locations"""
    locations = ddlpy.locations()
    return locations


@pytest.mark.unittest
def test_ddlpy_to_hatyan(locations):
    # input parameters
    tstart_dt = dt.datetime(2023,12,24)
    tstop_dt = dt.datetime(2024,1,5)

    bool_station = locations.index.isin(['HOEKVHLD'])
    bool_hoedanigheid = locations['Hoedanigheid.Code'].isin(['NAP'])
    bool_grootheid = locations['Grootheid.Code'].isin(['WATHTE'])
    bool_groepering = locations['Groepering.Code'].isin(['NVT'])
    locs_wathte = locations.loc[bool_station & bool_grootheid &
                                bool_groepering & bool_hoedanigheid]

    meas_wathte = ddlpy.measurements(locs_wathte.iloc[0], start_date=tstart_dt, end_date=tstop_dt)
    
    # timeseries
    ts_measwl = hatyan.ddlpy_to_hatyan(meas_wathte)

    assert ts_measwl.columns.tolist() == ['values', 'qualitycode', 'status']
    assert ts_measwl.index.name == 'time'

    assert ptypes.is_float_dtype(ts_measwl["values"])
    assert ptypes.is_integer_dtype(ts_measwl["qualitycode"])
    assert ptypes.is_object_dtype(ts_measwl["status"])


@pytest.mark.unittest
def test_convert_hwlwstr2num(locations):

    tstart_dt = "2022-12-19"
    tstop_dt = "2022-12-31"
    
    bool_grootheid_meas = locations['Grootheid.Code'].isin(['WATHTE'])
    bool_hoedanigheid = locations['Hoedanigheid.Code'].isin(['NAP'])
    bool_groepering_ext = locations['Groepering.Code'].isin(['GETETM2','GETETMSL2'])
    bool_grootheid_exttypes = locations['Grootheid.Code'].isin(['NVT'])
    bool_groepering_ext_meas = locations['Groepering.Code'].isin(['GETETM2','GETETMSL2'])
    bool_station = locations.index.isin(["HOEKVHLD"])
    locs_wathte_ext = locations.loc[bool_grootheid_meas & bool_hoedanigheid & bool_groepering_ext & bool_station]
    locs_exttypes_wathte = locations.loc[bool_grootheid_exttypes & bool_groepering_ext_meas & bool_station]
    
    # no support for multiple rows, so pass one at a time
    meas_wathte_ext = ddlpy.measurements(locs_wathte_ext.iloc[0], start_date=tstart_dt, end_date=tstop_dt)
    meas_wathte_exttypes = ddlpy.measurements(locs_exttypes_wathte.iloc[0], start_date=tstart_dt, end_date=tstop_dt)
    # hatyan timeseries
    ts_measwlHWLW = hatyan.ddlpy_to_hatyan(meas_wathte_ext, meas_wathte_exttypes)
    
    hwlwcode_expected = np.array([2, 1, 2, 1, 2, 1, 3, 4, 5, 1, 2, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4,
            5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1,
            3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4, 5, 1, 3, 4,
            5, 1, 2, 1, 3, 4, 5, 1, 2, 1, 2, 1])
    
    assert "HWLWcode" in ts_measwlHWLW.columns
    assert np.array_equal(ts_measwlHWLW["HWLWcode"].values, hwlwcode_expected)



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 11 14:33:40 2023

@author: veenstra
"""

import os
import pytest
import numpy as np
import hatyan
from hatyan.metadata import metadata_from_obj

dir_tests = os.path.dirname(__file__) #F9 doesnt work, only F5 (F5 also only method to reload external definition scripts)
dir_testdata = os.path.join(dir_tests,'data_unitsystemtests')


@pytest.mark.unittest
def test_read_write_components():
    
    current_station = 'HOEKVHLD'
    file_orig = os.path.join(dir_testdata,f'{current_station}_ana.txt')
    file_new = 'temp_components.txt'
    
    comp_orig = hatyan.read_components(filename=file_orig)
    hatyan.write_components(comp_orig, filename=file_new)
    comp_new = hatyan.read_components(filename=file_new)
    
    meta_orig = metadata_from_obj(comp_orig)
    meta_new = metadata_from_obj(comp_new)
    
    assert np.allclose(comp_orig, comp_new)
    assert meta_orig == meta_new
    os.remove(file_new)


@pytest.mark.unittest
def test_read_write_components_nometadata():
    """
    Test for component files written with hatyan 2.7.0 or lower,
    these files lack essential metadata for STAT, PERD and CODE lines.
    This tests tests whether the correct exception is raised.
    """
    
    file_orig = os.path.join(dir_testdata,'DENHDR_ana.txt')
    file_nometa = "temp_comp_dummymetadata.txt"
    with open(file_orig, "r") as f:
        data = f.readlines()
    with open(file_nometa, "w") as f:
        for line in data:
            if line.startswith("STAT"):
                continue
            if line.startswith("PERD"):
                continue
            if line.startswith("CODE"):
                continue
            f.write(line)
    
    try:
        _ = hatyan.read_components(file_nometa)
    except KeyError as e:
        assert "No STAT/PERD metadata available in component file" in str(e)

    os.remove(file_nometa)


@pytest.mark.unittest
def test_read_write_components_nondefaultsettings():
    
    current_station = 'HOEKVHLD'
    file_orig = os.path.join(dir_testdata,f'{current_station}_ana.txt')
    file_new = 'temp_components.txt'
    
    comp_orig = hatyan.read_components(filename=file_orig)
    comp_orig.attrs["nodalfactors"] = False
    with pytest.raises(ValueError) as e:
        hatyan.write_components(comp_orig, filename=file_new)
    assert "nodalfactors" in str(e.value)
    
    comp_orig = hatyan.read_components(filename=file_orig)
    comp_orig.attrs["fu_alltimes"] = True
    with pytest.raises(ValueError) as e:
        hatyan.write_components(comp_orig, filename=file_new)
    assert "fu_alltimes" in str(e.value)
    
    comp_orig = hatyan.read_components(filename=file_orig)
    comp_orig.attrs["source"] = "notschureman"
    with pytest.raises(ValueError) as e:
        hatyan.write_components(comp_orig, filename=file_new)
    assert "source" in str(e.value)

    comp_orig = hatyan.read_components(filename=file_orig)
    comp_orig.attrs['tzone'] = None
    with pytest.raises(ValueError) as e:
        hatyan.write_components(comp_orig, filename=file_new)
    assert "tzone=None" in str(e.value)


@pytest.mark.unittest
def test_writecomponents_fromanalysis(tmp_path):
    """
    this is of added value to check if all required metadata is present from an analysis
    """
    current_station = 'VLISSGN'
    file_data_comp0 = os.path.join(dir_testdata,f'{current_station}_obs1.txt')
    ts = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    
    comp = hatyan.analysis(ts=ts, const_list='month', fu_alltimes=False)
    file_comp = os.path.join(tmp_path, "temp_comp.txt")
    hatyan.write_components(comp, file_comp)


@pytest.mark.unittest
def test_plot_components_validation():
    
    current_station = 'HOEKVHLD'
    file_data_comp0 = os.path.join(dir_testdata,f'{current_station}_ana.txt')
    
    comp_merged = hatyan.read_components(filename=file_data_comp0)
    _ = hatyan.plot_components(comp=comp_merged, comp_validation=comp_merged)


@pytest.mark.unittest
def test_plot_components_allyears():
    current_station = 'VLISSGN'
    file_data_comp0 = os.path.join(dir_testdata,f'{current_station}_obs?.txt')
    ts_measurements = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    
    ts_comp, ts_comp_all = hatyan.analysis(ts=ts_measurements, const_list='month', analysis_perperiod="Y", return_allperiods=True)
    _ = hatyan.plot_components(comp=ts_comp, comp_allperiods=ts_comp_all)


@pytest.mark.unittest
def test_merge_componentgroups():
    file_comp1 = os.path.join(dir_testdata,'VLISSGN_ana.txt')
    comp_fromfile1 = hatyan.read_components(filename=file_comp1)
    file_comp2 = os.path.join(dir_testdata,'HOEKVHLD_ana.txt')
    comp_fromfile2 = hatyan.read_components(filename=file_comp2)
    comp_fromfile2.attrs["station"] = "VLISSGN"
    
    comp_merged = hatyan.merge_componentgroups(comp_main=comp_fromfile1, comp_sec=comp_fromfile2.loc[['SA','SM']])
    
    assert (comp_fromfile1!=comp_merged).sum().sum() == 4


@pytest.mark.unittest
def test_merge_componentgroups_comparesettings():
    file_data_comp = os.path.join(dir_testdata,'VLISSGN_ana.txt')
    comp_fromfile = hatyan.read_components(filename=file_data_comp)
    
    with pytest.raises(ValueError):
        comp_fromfile_fake = comp_fromfile.copy()
        comp_fromfile_fake.attrs["nodalfactors"] = False
        _ = hatyan.merge_componentgroups(comp_main=comp_fromfile, comp_sec=comp_fromfile_fake.loc[['SA','SM']])

    with pytest.raises(ValueError):
        comp_fromfile_fake = comp_fromfile.copy()
        comp_fromfile_fake.attrs["xfac"] = False
        _ = hatyan.merge_componentgroups(comp_main=comp_fromfile, comp_sec=comp_fromfile_fake.loc[['SA','SM']])

    with pytest.raises(ValueError):
        comp_fromfile_fake = comp_fromfile.copy()
        comp_fromfile_fake.attrs["fu_alltimes"] = True
        _ = hatyan.merge_componentgroups(comp_main=comp_fromfile, comp_sec=comp_fromfile_fake.loc[['SA','SM']])

    with pytest.raises(ValueError):
        comp_fromfile_fake = comp_fromfile.copy()
        comp_fromfile_fake.attrs["source"] = 'foreman'
        _ = hatyan.merge_componentgroups(comp_main=comp_fromfile, comp_sec=comp_fromfile_fake.loc[['SA','SM']])



# -*- coding: utf-8 -*-
"""
analysis_prediction.py contains hatyan definitions related to tidal analysis and prediction. 
"""

import numpy as np
import pandas as pd
import datetime as dt
import logging

from hatyan.hatyan_core import get_const_list_hatyan, sort_const_list, robust_timedelta_sec
from hatyan.hatyan_core import get_freqv0_generic, get_uf_generic
from hatyan.timeseries import Timeseries_Statistics, nyquist_folding, check_rayleigh
from hatyan.metadata import metadata_from_obj, metadata_add_to_obj
from hatyan.deprecated import (deprecated_python_option,
                               DEPRECATED_OPTIONS_PREDICTION_DICT,
                               DEPRECATED_OPTIONS_ANALYSIS_DICT)

__all__ = ["analysis",
           "prediction",
           ]

logger = logging.getLogger(__name__)


class PydanticConfig:
    #https://docs.pydantic.dev/usage/model_config/
    arbitrary_types_allowed = True #necessary to allow for pd.DataFrame etc


class MatrixConditionTooHigh(Exception):
    pass


class HatyanSettings:
    """
    Settings class containing default hatyan settings, to be overwritten by input, initiate with:
    hatyan_settings = hatyan.HatyanSettings(nodalfactors=False)

    """
    
    def __init__(self, 
                 nodalfactors, fu_alltimes, xfac, source, #prediction/analysis 
                 cs_comps=None, analysis_perperiod=None, return_allperiods=None, #analysis only
                 max_matrix_condition=None): #analysis only
        
        if not isinstance(nodalfactors,bool):
            raise TypeError(f'invalid nodalfactors={nodalfactors} type, should be bool')
        self.nodalfactors = nodalfactors
        
        if not isinstance(fu_alltimes,bool):
            raise TypeError(f'invalid fu_alltimes={fu_alltimes} type, should be bool')
        self.fu_alltimes = fu_alltimes
        
        if not (isinstance(xfac,bool) or isinstance(xfac,dict)):
            raise TypeError(f'invalid xfac={xfac} type, should be bool or dict')
        self.xfac = xfac
        
        if not isinstance(source,str):
            raise TypeError(f'invalid source={source} type, should be str')
        source = source.lower()
        if source not in ['schureman','foreman']:
            raise TypeError(f'invalid source {source}, should be "schureman" or "foreman"')
        self.source = source
                
        if return_allperiods is not None:
            if not isinstance(return_allperiods,bool):
                raise TypeError(f'invalid return_allperiods={return_allperiods} type, should be bool')
            self.return_allperiods = return_allperiods
        
        if analysis_perperiod is not None:
            if not ((analysis_perperiod is False) or (analysis_perperiod in ['Y','Q','M'])):
                raise TypeError(f'invalid analysis_perperiod={analysis_perperiod} type, should be False or Y/Q/M')
            self.analysis_perperiod = analysis_perperiod
        
        if cs_comps is not None:
            if not isinstance(cs_comps,(dict,pd.DataFrame)):
                raise TypeError('invalid cs_comps type, should be dict')
            cs_comps = pd.DataFrame(cs_comps) #TODO: convert all to dict or pd.DataFrame
            cs_comps_expectedkeys = ['CS_comps_derive', 'CS_comps_from', 'CS_ampfacs', 'CS_degincrs']
            for cs_comps_key in cs_comps_expectedkeys:
                if cs_comps_key not in cs_comps.keys():
                    raise KeyError(f'cs_comps does not contain {cs_comps_key}')
            cs_comps_lenvals = [len(cs_comps[key]) for key in cs_comps]
            if len(np.unique(cs_comps_lenvals)) != 1:
                raise ValueError(f'cs_comps keys do not have equal lengths:\n{cs_comps}')
            self.cs_comps = cs_comps
        
        if max_matrix_condition is not None:
            if not isinstance(max_matrix_condition,int) or isinstance(max_matrix_condition,float):
                raise TypeError(f'invalid {max_matrix_condition} type, should be int or float')
            self.max_matrix_condition = max_matrix_condition
        
    def __str__(self):
        self_dict = vars(self)
        str_append = ''
        for key,val in self_dict.items():
            if key=='cs_comps':
                str_append += f'{key:20s} = \n{val}\n'
            else:
                str_append += f'{key:20s} = {val}\n'
        return str_append


def vectoravg(A_all, phi_deg_all):
    """
    calculates the vector average of A and phi per constituent, 
    it vector averages over values resulting from multiple periods.
    A regular average is calculated for the amplitude of A0 (middenstand)
    
    Parameters
    ----------
    A_i_all : TYPE
        DESCRIPTION.
    phi_i_deg_all : TYPE
        DESCRIPTION.

    Returns
    -------
    A_i_mean : TYPE
        DESCRIPTION.
    phi_i_deg_mean : TYPE
        DESCRIPTION.

    """
        
    phi_rad_all = np.deg2rad(phi_deg_all)
    v_cos = A_all*np.cos(phi_rad_all)
    v_sin = A_all*np.sin(phi_rad_all)
    mean_v_cos = np.mean(v_cos,axis=1)
    mean_v_sin = np.mean(v_sin,axis=1)
    A_mean = np.sqrt(mean_v_cos**2 + mean_v_sin**2)
    phi_rad_mean = np.arctan2(mean_v_sin,mean_v_cos)
    phi_rad_mean[phi_rad_mean<0] = phi_rad_mean[phi_rad_mean<0]+(2*np.pi)
    
    #if phases of all years are exactly 0, it is the A0 component. Overwrite this A0 with mean amplitude and zero phase if present, otherwise negative values will become positive with 180 phase
    idx_A0 = np.nonzero((phi_deg_all==0).any(axis=1))[0]
    A_mean[idx_A0] = np.mean(A_all[idx_A0,:])
    phi_rad_mean[idx_A0] = 0
    
    phi_deg_mean = np.rad2deg(phi_rad_mean)
    
    return A_mean, phi_deg_mean


@deprecated_python_option(**DEPRECATED_OPTIONS_ANALYSIS_DICT)
def analysis(ts, const_list, 
             nodalfactors=True, fu_alltimes=True, xfac=False, 
             source='schureman',
             cs_comps=None,
             analysis_perperiod=False, return_allperiods=False,
             max_matrix_condition=12):
    """
    Analysis of timeseries.
    Optionally processes a timeseries per year and vector averages the results afterwards.
    The timezone of the timeseries, will also be reflected in the phases of the resulting component set, 
    so the resulting component set can be used to make a prediction in the original timezone.
    
    Parameters
    ----------
    ts : pandas.DataFrame
        The DataFrame should contain a 'values' column and a pd.DatetimeIndex as index, it contains the timeseries to be analysed, as obtained from e.g. readts_*.
    const_list : list, pandas.Series or str
        list or pandas.Series: contains the tidal constituent names for which to analyse the provided timeseries ts. 
        str: a predefined name of a component set for hatyan_core.get_const_list_hatyan()
    nodalfactors : bool
        Whether or not to apply nodal factors. The default is True.
    fu_alltimes : bool
        Whether to calculate nodal factors in middle of the analysis/prediction period (default) or on every timestep. The default is True.
    xfac : bool
        Whether or not to apply x-factors. The default is False.
    source : TYPE
        DESCRIPTION. The default is 'schureman'.
    cs_comps : pandas.DataFrame, optional
        contains the from/derive component lists for components splitting, as well as the amplitude factor and the increase in degrees. Only relevant for analysis. The default is None.
    max_matrix_condition: float or int
        the maximum condition of the xTx matrix. The default is 12.
    analysis_perperiod : False or Y/Q/W, optional
        caution, it tries to analyse each year/quarter/month, but skips if it fails. The default is False.
    return_allperiods : bool, optional
        Only relevant if analysis_perperiod is not None. The default is False.
    

    Returns
    -------
    COMP_mean_pd : pandas.DataFrame
        The DataFrame contains the component data with component names as index, and colums 'A' and 'phi_deg'.
    COMP_all_pd : pandas.DataFrame, optional
        The same as COMP_mean_pd, but with all years added with MultiIndex
    """
    # check ts.index type, this works better than isinstance(ts.index,pd.DatetimeIndex) since 1018 time indexes are Index instead of DatetimeIndex
    if not (isinstance(ts.index[0],pd.Timestamp) or isinstance(ts.index[0],dt.datetime)):
        raise TypeError(f'ts.index is not of expected type ({type(ts.index[0])} instead of pd.Timestamp or dt.datetime)')
    
    # remove timezone from timeseries (is added to components dataframe after analysis)
    ts_pd = ts.copy()
    ts_pd.index = ts_pd.index.tz_localize(None)
    
    # validate settings
    hatyan_settings = HatyanSettings(source=source, nodalfactors=nodalfactors, fu_alltimes=fu_alltimes, xfac=xfac,
                                     analysis_perperiod=analysis_perperiod, return_allperiods=return_allperiods,
                                     cs_comps=cs_comps, max_matrix_condition=max_matrix_condition)
    
    logger.info(f'ANALYSIS initializing\n{hatyan_settings}')
        
    #retrieving and sorting const_list
    if type(const_list) is str:
        const_list = get_const_list_hatyan(const_list)
    const_list = sort_const_list(const_list)
    logger.info(f'n components analyzed  = {len(const_list)}')
    
    #check for duplicate components (results in singular matrix)
    if len(const_list) != len(np.unique(const_list)):
        const_list_uniq, const_list_uniq_counts = np.unique(const_list,return_counts=True)
        const_list_counts = pd.DataFrame({'constituent':const_list_uniq,'occurences':const_list_uniq_counts})
        raise ValueError('remove duplicate constituents from const_list:\n%s'%(const_list_counts.loc[const_list_counts['occurences']>1]))
    
    n_const = len(const_list)
    if hasattr(hatyan_settings, "cs_comps"):
        n_const = len(const_list) + len(hatyan_settings.cs_comps)

    if hatyan_settings.analysis_perperiod:
        period = hatyan_settings.analysis_perperiod
        logger.info(f'analysis_perperiod={period}, separate periods are automatically determined from timeseries')
        ts_periods_dt_all = ts_pd.index.to_period(period)
        ts_periods_dt = ts_periods_dt_all.unique() # TODO: to_period is not limited to Y/Q/M, there are more options that are now blocked: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        ts_periods_strlist = [str(x) for x in ts_periods_dt]
        
        n_periods = len(ts_periods_dt)
        A_i_all = np.zeros((n_const,n_periods))*np.nan
        phi_i_deg_all = np.zeros((n_const,n_periods))*np.nan
        for iP, period_dt in enumerate(ts_periods_dt):
            logger.info('analyzing %s of sequence %s'%(period_dt,ts_periods_strlist))
            ts_oneperiod_pd = ts_pd[ts_periods_dt_all==period_dt]
            try:
                COMP_one = analysis_singleperiod(ts_oneperiod_pd, const_list=const_list, hatyan_settings=hatyan_settings)
                A_i_all[:,iP] = COMP_one.loc[:,'A']
                phi_i_deg_all[:,iP] = COMP_one.loc[:,'phi_deg']
            except MatrixConditionTooHigh: # accept exception if matrix condition is too high, since some years can then be skipped
                logger.warning(f'analysis of {period_dt} failed because MatrixConditionTooHigh, check if const_list is appropriate for timeseries lenght.')
        if np.isnan(A_i_all).all():
            raise ValueError('all nans: analysis perperiod failed for all periods, check warnings above')
        
        COMP_all_pd = pd.DataFrame(data=np.hstack([A_i_all,phi_i_deg_all]), columns=pd.MultiIndex.from_product([['A','phi_deg'],ts_periods_dt]), index=COMP_one.index)
        logger.info('vector averaging analysis results')
        A_i_mean, phi_i_deg_mean = vectoravg(A_all=A_i_all, phi_deg_all=phi_i_deg_all)
        COMP_mean_pd = pd.DataFrame({ 'A': A_i_mean, 'phi_deg': phi_i_deg_mean},index=COMP_one.index)
    else: #dummy values, COMP_years should be equal to COMP_mean
        COMP_mean_pd = analysis_singleperiod(ts_pd, const_list=const_list, hatyan_settings=hatyan_settings)
        COMP_all_pd = None
    
    #add metadata
    metadata = metadata_from_obj(ts_pd)
    metadata['nodalfactors'] = hatyan_settings.nodalfactors
    metadata['xfac'] = hatyan_settings.xfac
    metadata['fu_alltimes'] = hatyan_settings.fu_alltimes
    metadata['source'] = hatyan_settings.source
    metadata['tstart'] = ts.index.min().tz_localize(None)
    metadata['tstop'] = ts.index.max().tz_localize(None)
    metadata['tzone'] = ts.index.tz
    COMP_mean_pd = metadata_add_to_obj(COMP_mean_pd, metadata)
    
    if hatyan_settings.return_allperiods:
        COMP_all_pd = metadata_add_to_obj(COMP_all_pd, metadata)
        return COMP_mean_pd, COMP_all_pd
    
    logger.info('ANALYSIS finished')
    
    return COMP_mean_pd


def analysis_singleperiod(ts, const_list, hatyan_settings):
    """
    harmonic analysis with matrix transformations (least squares fit), optionally with component splitting
    for details about arguments and return variables, see analysis() definition
    
    """
    
    #drop duplicate times
    bool_ts_duplicated = ts.index.duplicated(keep='first')
    ts_pd = ts.copy() #TODO: this is not necessary
    if bool_ts_duplicated.any():
        raise ValueError(f'{bool_ts_duplicated.sum()} duplicate timesteps in provided timeseries, remove them e.g. with: ts = ts[~ts.index.duplicated(keep="first")]')
    message = (f'#timesteps    = {len(ts)}\n'
               f'tstart        = {ts.index.min().strftime("%Y-%m-%d %H:%M:%S")}\n'
               f'tstop         = {ts.index.max().strftime("%Y-%m-%d %H:%M:%S")}\n'
               f'timestep      = {ts.index.freq}')
    logger.info(message)

    #remove nans
    ts_pd_nonan = ts_pd[~ts_pd['values'].isna()]
    if len(ts_pd_nonan)<2:
        raise ValueError(f'provided timeseries is less than 2 timesteps long (after dropping potential nans), analysis not possible:\n{ts_pd_nonan}')
    times_pred_all_pdDTI = ts_pd_nonan.index.copy() #pd.DatetimeIndex(ts_pd_nonan.index) #TODO: this will not work for OutOfBoundsDatetime
    percentage_nan = 100-len(ts_pd_nonan['values'])/len(ts_pd['values'])*100
    logger.info(f'percentage_nan in values_meas_sel: {percentage_nan:.2f}%')
    
    #get times and time array
    dood_date_mid = ts_pd.index[[len(ts_pd.index)//2]] #middle of analysis period (2july in case of 1jan-1jan), zoals bij hatyan #TODO: this is incorrect in case of e.g. more missings in first half of year than second half
    dood_date_start = ts_pd.index[[0]] #first date (for v0, also freq?)
    if hatyan_settings.fu_alltimes:
        dood_date_fu = times_pred_all_pdDTI
    else:
        dood_date_fu = dood_date_mid
    times_from0_s = robust_timedelta_sec(ts_pd_nonan.index,refdate_dt=dood_date_start[0])
    times_from0_s = times_from0_s[:,np.newaxis]
    
    #get frequency and v0
    t_const_freq_pd, v_0i_rad = get_freqv0_generic(const_list, dood_date_mid, dood_date_start, hatyan_settings.source)
    omega_i_rads = t_const_freq_pd[['freq']].values.T*(2*np.pi)/3600 #angular frequency, 2pi/T, in rad/s, https://en.wikipedia.org/wiki/Angular_frequency (2*np.pi)/(1/x*3600) = 2*np.pi*x/3600
    u_i_rad, f_i = get_uf_generic(const_list, dood_date_fu, hatyan_settings.nodalfactors, hatyan_settings.xfac, hatyan_settings.source)
    v_u = v_0i_rad.values + u_i_rad.values
    
    #check rayleigh frequency after nyquist frequency folding process.
    freq_rem = nyquist_folding(ts_pd,t_const_freq_pd)
    check_rayleigh(ts_pd,freq_rem) #TODO: maybe sometimes valuable to not fold with nyquist (eg with strongly varying time interval), in that case: check_rayleigh(ts_pd,t_const_freq_pd)

    #### TIMESERIES ANALYSIS
    N = len(const_list)
    m = len(ts_pd_nonan['values'])
    
    # get xmat and make dot product
    xmat = np.zeros((m,2*N))
    xmat[:,:N] = f_i.values * np.cos(omega_i_rads*times_from0_s+v_u)
    xmat[:,N:] = f_i.values * np.sin(omega_i_rads*times_from0_s+v_u)
    
    xTmat = xmat.T
    logger.info('calculating xTx matrix')
    tic = dt.datetime.now()
    xTxmat = np.dot(xTmat,xmat)
    
    if 'A0' in const_list: #correct center value [N,N] for better matrix condition
        xTxmat_condition = np.linalg.cond(xTxmat)
        logger.debug('condition of xTx matrix before center adjustment for A0: %.2f'%(xTxmat_condition))
        xTxmat[N,N] = m
    xTxmat_condition = np.linalg.cond(xTxmat)
    logger.info('condition of xTx matrix: %.2f'%(xTxmat_condition))
    if xTxmat_condition > hatyan_settings.max_matrix_condition:#10:#100: #random treshold
        raise MatrixConditionTooHigh(f'ERROR: condition of xTx matrix is too high ({xTxmat_condition:.2f}), check your timeseries length, try different (shorter) component set or componentsplitting.\nAnalysed {Timeseries_Statistics(ts_pd)}')
    xTymat = np.dot(xTmat,ts_pd_nonan['values'].values)
    
    #solve matrix to get beta_roof_mat (and thus a, b)
    beta_roof_mat = np.linalg.solve(xTxmat,xTymat)
    logger.info('matrix system solved, elapsed time: %s'%(dt.datetime.now()-tic))

    phi_i_rad = np.arctan2(beta_roof_mat[N:],beta_roof_mat[:N]) #(a,b) arctan_ab
    A_i = np.sqrt(beta_roof_mat[N:]**2 + beta_roof_mat[:N]**2) #(a,b) sqsqrt_ab

    COMP_pd = pd.DataFrame({'A': A_i, 'phi_deg': np.rad2deg(phi_i_rad%(2*np.pi))}, index=const_list)
    if 'A0' in COMP_pd.index: #correct 180 degrees A0 phase by making amplitude value negative
        if COMP_pd.loc['A0','phi_deg']==180:
            COMP_pd.loc['A0','A'] = -COMP_pd.loc['A0','A']
            COMP_pd.loc['A0','phi_deg'] = 0
    
    if hasattr(hatyan_settings, "cs_comps"):
        COMP_pd = split_components(comp=COMP_pd, dood_date_mid=dood_date_mid, hatyan_settings=hatyan_settings)
        
    return COMP_pd


def split_components(comp, dood_date_mid, hatyan_settings):
    """
    component splitting function
    for details about arguments and return variables, see analysis() definition

    """
    
    for cs_compname in hatyan_settings.cs_comps['CS_comps_derive']:
        if cs_compname in comp.index:
            raise ValueError(f"component {cs_compname} requested via component splitting, but already present in const_list")
    
    #create sorted and complete component list
    const_list_inclcs_raw = comp.index.tolist() + hatyan_settings.cs_comps['CS_comps_derive'].tolist()
    const_list_inclcs = sort_const_list(const_list=const_list_inclcs_raw)

    #retrieve freq and speed
    _, cs_v_0i_rad = get_freqv0_generic(const_list=const_list_inclcs, dood_date_mid=dood_date_mid, dood_date_start=dood_date_mid, source=hatyan_settings.source) # with split_components, v0 is calculated on the same timestep as u and f (middle of original series)
    cs_u_i_rad, cs_f_i = get_uf_generic(const_list=const_list_inclcs, dood_date_fu=dood_date_mid, nodalfactors=hatyan_settings.nodalfactors, xfac=hatyan_settings.xfac, source=hatyan_settings.source)
    
    comp_inclcs = pd.DataFrame(comp,index=const_list_inclcs,columns=comp.columns)
        
    for comp_main in np.unique(hatyan_settings.cs_comps['CS_comps_from']):
        # boolean of which rows of cs_comps dataframe corresponds to a main constituent, also makes it possible to select two rows
        bool_cs_maincomp = hatyan_settings.cs_comps['CS_comps_from'] == comp_main
        cs_comps_formain = hatyan_settings.cs_comps.loc[bool_cs_maincomp]
        comp_slave_list = cs_comps_formain['CS_comps_derive'].tolist()
        logger.info(f'splitting component {comp_main} into {comp_slave_list}')
        
        #first update main components based on degincrs/ampfacs of all components that are to be derived
        DBETA = 0
        for iR, cs_comps_row in cs_comps_formain.iterrows():
            comp_slave = cs_comps_row['CS_comps_derive']
            #code from resuda.f, line 440 to 455
            DMU = cs_f_i.loc[0,comp_slave]/cs_f_i.loc[0,comp_main]
            DTHETA = cs_comps_row['CS_ampfacs']
            DGAMMA = (np.deg2rad(cs_comps_row['CS_degincrs']) - 
                      DBETA - 
                      (cs_v_0i_rad.loc[0,comp_slave] + cs_u_i_rad.loc[0,comp_slave]) + 
                      (cs_v_0i_rad.loc[0,comp_main] + cs_u_i_rad.loc[0,comp_main])) #in FORTRAN code, CS_f_i slave/main is also added, this seems wrong
            DREEEL = 1+DMU*DTHETA*np.cos(DGAMMA)
            DIMAGI = DMU*DTHETA*np.sin(DGAMMA)  
            DALPHA = np.sqrt(DREEEL*DREEEL+DIMAGI*DIMAGI)
            if DALPHA < 1e-50:
                raise Exception('ERROR: DALPHA too small, component splitting failed?')
            DBETA = np.arctan2(DIMAGI,DREEEL)
            
            comp_inclcs.loc[comp_main,'A'] = comp_inclcs.loc[comp_main,'A']/DALPHA
            comp_inclcs.loc[comp_main,'phi_deg'] = (comp_inclcs.loc[comp_main,'phi_deg']-np.rad2deg(DBETA))%360 
        
        #updating slave components after updating main components, this makes a difference when splitting a component into more than two
        for iR, cs_comps_row in cs_comps_formain.iterrows():
            comp_slave = cs_comps_row['CS_comps_derive']
            comp_inclcs.loc[comp_slave,'A'] = comp_inclcs.loc[comp_main,'A'] * cs_comps_row['CS_ampfacs']
            comp_inclcs.loc[comp_slave,'phi_deg'] = (comp_inclcs.loc[comp_main,'phi_deg'] + cs_comps_row['CS_degincrs'])%360
            
    return comp_inclcs


def prediction_singleperiod(comp:pd.DataFrame, times:pd.DatetimeIndex, hatyan_settings) -> pd.DataFrame:
    
    metadata_comp = metadata_from_obj(comp)
    tzone_comp = metadata_comp.pop('tzone', None)
    
    tzone_pred = times.tz
    
    if tzone_pred is None and tzone_comp is None:
        tzone_convert = False
    elif tzone_pred is not None and tzone_comp is not None:
        tzone_convert = True
    else:
        raise ValueError("provided times and components should both be timezone-aware "
                         "or timezone-naive, not mixed.")
    
    # remove timezone from prediction times: first convert times to tzone of components, then make timezone naive
    if tzone_convert:
        times = times.tz_convert(tzone_comp)
        times = times.tz_localize(None)
        # TODO: workaround to maintain the frequency dropped by tz_localize
        # https://github.com/pandas-dev/pandas/issues/36575
        times.freq = times.inferred_freq
    
    logger.info(f'components used = {len(comp)}\n'
                f'tstart = {times[0].strftime("%Y-%m-%d %H:%M:%S")}\n'
                f'tstop = {times[-1].strftime("%Y-%m-%d %H:%M:%S")}\n'
                f'timestep = {times.freq}')
    
    # middle of analysis period (2july in case of 1jan-1jan), zoals bij hatyan.
    dood_date_mid = times[[len(times)//2]]
    # first date (for v0, also freq?)
    dood_date_start = times[:1]
    
    # sort component list and component dataframe
    if np.isnan(comp.values).any():
        raise Exception('provided component set contains nan values, prediction not possible')
    const_list = sort_const_list(comp.index.tolist())
    COMP = comp.loc[const_list]
    A = np.array(COMP['A'])
    phi_rad = np.array(np.deg2rad(COMP['phi_deg']))

    t_const_freq_pd, v_0i_rad = get_freqv0_generic(const_list, dood_date_mid, dood_date_start, hatyan_settings.source)
    t_const_speed_all = t_const_freq_pd['freq'].values[:,np.newaxis]*(2*np.pi)

    if hatyan_settings.fu_alltimes:
        dood_date_fu = times
    else:
        dood_date_fu = dood_date_mid
    u_i_rad, f_i = get_uf_generic(const_list, dood_date_fu, hatyan_settings.nodalfactors, hatyan_settings.xfac, hatyan_settings.source)

    logger.info('PREDICTION started')
    omega_i_rads = t_const_speed_all.T/3600 #angular frequency, 2pi/T, in rad/s, https://en.wikipedia.org/wiki/Angular_frequency (2*np.pi)/(1/x*3600) = 2*np.pi*x/3600
    if not isinstance(times,pd.DatetimeIndex): #support for years<1677, have to use Index instead of DatetimeIndex (DatetimeIndex is also Index, so isinstance(times_pred_all_pdDTI,pd.Index) does not work
        tdiff = pd.TimedeltaIndex(times-dood_date_start) #pd.TimedeltaIndex is around it to avoid it being an Index in case of outofbounds timesteps (necessary from pandas 2.0.0)
    else:
        tdiff = pd.TimedeltaIndex(times-dood_date_start[0]) #pd.TimedeltaIndex is not necessary here, but for conformity with above
    times_from0allpred_s_orig = tdiff.total_seconds().values
    times_from0allpred_s = np.transpose(times_from0allpred_s_orig[np.newaxis])
    
    f_A = np.multiply(f_i.values,A)
    omeg_t = np.multiply(times_from0allpred_s,omega_i_rads)
    v_u_phi = np.subtract(np.add(v_0i_rad.values,u_i_rad.values),phi_rad)
    omeg_t_v_u_phi = np.add(omeg_t,v_u_phi)
    ht_res = np.sum(np.multiply(f_A,np.cos(omeg_t_v_u_phi)),axis=1) #not necessary to add A0, since it is already part of the component list
    
    ts_prediction_pd = pd.DataFrame({'values': ht_res}, index=times)
    logger.info('PREDICTION finished')
    
    # add timezone to prediction: first interpret times as tzone of components, then convert to timezone of prediction
    if tzone_convert:
        ts_prediction_pd = ts_prediction_pd.tz_localize(tzone_comp)
        # TODO: workaround to maintain the frequency dropped by tz_localize
        # https://github.com/pandas-dev/pandas/issues/36575
        ts_prediction_pd.index.freq = ts_prediction_pd.index.inferred_freq
        ts_prediction_pd = ts_prediction_pd.tz_convert(tzone_pred)
    
    return ts_prediction_pd


@deprecated_python_option(**DEPRECATED_OPTIONS_PREDICTION_DICT)
def prediction(comp, times=None, timestep=None):
    """
    generates a tidal prediction from a set of components A and phi values.
    The component set has the same timezone as the timeseries used to create it.
    If times is timezone-naive the resulting prediction will be in component timezone.
    If times is timezone-aware the resulting prediction will be converted to that timezone.
    If a components dataframe contains multiple column levels (multiple periods),
    The prediction is a concatenation of predictions of all periods (based on the respective A/phi values).

    Parameters
    ----------
    comp : pd.DataFrame
        The DataFrame contains the component data with component names as index, and colums 'A' and 'phi_deg'.
    times : (pd.DatetimeIndex,slice), optional
        pd.DatetimeIndex with prediction timeseries or slice(tstart,stop,timestep) to construct it from. 
        If None, pd.DatetimeIndex is constructed from the tstart/tstop/timestep metadata attrs of the comp object. 
        Only allowed/relevant for component dataframes with single-level columns (single period). The default is None.
    timestep : str
        Only allowed/relevant for component dataframes with multi-level columns (different periods). 
        The string is parsed with pandas.tseries.frequencies.to_offset(). The default is None.

    Returns
    -------
    ts_prediction : TYPE
        The DataFrame contains a 'values' column and a pd.DatetimeIndex as index, it contains the prediction times and values.

    """
    
    # get settings from component attribute and validate their values
    settings_kwargs = {}
    for setting in ['nodalfactors', 'xfac', 'fu_alltimes', 'source']:
        settings_kwargs[setting] = comp.attrs[setting]
    hatyan_settings = HatyanSettings(**settings_kwargs)
    
    logger.info(f'PREDICTION initializing\n{hatyan_settings}')
    
    metadata_comp = metadata_from_obj(comp)
    tzone_comp = metadata_comp.pop('tzone', None)
    
    if hasattr(comp.columns,"levels"):
        logger.info('prediction() per period due to levels in component dataframe columns')
        if timestep is None:
            raise TypeError("prediction() per period, so 'timestep' argument should not be None")
        if times is not None:
            raise TypeError("prediction() per period, so 'times' argument not allowed")
        # convert timestep to tstep of proper type
        tstep = pd.tseries.frequencies.to_offset(timestep)
        
        ts_periods_dt = comp.columns.levels[1]
        ts_periods_strlist = [str(x) for x in ts_periods_dt]
        
        ts_prediction_perperiod_list = []
        for period_dt in ts_periods_dt:
            logger.info(f'generating prediction {period_dt} of sequence {ts_periods_strlist}')
            comp_oneyear = comp.loc[:,(slice(None),period_dt)]
            comp_oneyear.columns = comp_oneyear.columns.droplevel(1)
            if period_dt.freqstr in ['A-DEC','Y-DEC']: #year frequency
                tstart = pd.Timestamp(period_dt.year, 1, 1)
                tstop = pd.Timestamp(period_dt.year+1, 1, 1) - pd.Timedelta(tstep)
            elif period_dt.freqstr in ['M']: #month frequency
                tstart = period_dt.to_timestamp()
                tstop = period_dt.to_timestamp() + pd.Timedelta(days=period_dt.days_in_month) - pd.Timedelta(tstep)
            else:
                raise Exception(f'unknown freqstr: {period_dt.freqstr}')
            # generate date range and do prediction
            times_pred = pd.date_range(start=tstart, end=tstop, freq=tstep, unit="us", tz=tzone_comp)
            ts_prediction_oneperiod = prediction_singleperiod(comp=comp_oneyear, times=times_pred, hatyan_settings=hatyan_settings)
            ts_prediction_perperiod_list.append(ts_prediction_oneperiod)
        ts_prediction = pd.concat(ts_prediction_perperiod_list)
    else:
        logger.info('prediction() atonce')
        if timestep is not None:
            raise TypeError("prediction() atonce, so 'timestep' argument not allowed")
        if times is None:
            raise TypeError("prediction() atonce, so 'times' argument should not be None")
        if isinstance(times,slice):
            tstart = pd.Timestamp(times.start)
            tstop = pd.Timestamp(times.stop)
            tstep = pd.tseries.frequencies.to_offset(times.step)
            times = pd.date_range(start=tstart, end=tstop, freq=tstep, unit="us")
        if not isinstance(times, pd.DatetimeIndex):
            raise TypeError(f'times argument can be of type pd.DatetimeIndex or slice, not {type(times)}')

        # backwards compatibility for timezone-naive times
        tzone_pred = times.tz
        if tzone_pred is None and tzone_comp is not None:
            times = times.tz_localize(tzone_comp)
            # TODO: workaround to maintain the frequency dropped by tz_localize
            # https://github.com/pandas-dev/pandas/issues/36575
            times.freq = times.inferred_freq
            logger.warning("provided times are timezone-naive and provided components are "
                            "timezone-aware. The times are being interpreted as if they would "
                            f"have the same timezone as the components: {tzone_comp}")
        
        ts_prediction = prediction_singleperiod(comp=comp, times=times, hatyan_settings=hatyan_settings)
    
    # add metadata (and update grootheid)
    metadata_comp = metadata_from_obj(comp)
    if 'grootheid' in metadata_comp:
        # update metadata
        if metadata_comp['grootheid'] == 'WATHTE':
            metadata_comp['grootheid'] = 'WATHTBRKD'
    # prevent adding time metadata from component dataframe to prediction dataframe
    if 'tzone' in metadata_comp.keys():
        metadata_comp.pop('tzone')
    if 'tstart' in metadata_comp.keys():
        metadata_comp.pop("tstart")
    if 'tstop' in metadata_comp.keys():
        metadata_comp.pop("tstop")
    ts_prediction = metadata_add_to_obj(ts_prediction, metadata_comp)

    return ts_prediction
